"""
Knowledge Builder - Create knowledge bases from various sources.

Supports:
- Web search and scraping
- API data ingestion
- Documents (PDF, DOCX, TXT, MD, etc.)
- Images (OCR and image analysis)
- Any file types
- Embedding generation for vector databases
"""

import os
import re
import json
import uuid
import logging
import mimetypes
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Callable, Dict, Iterator, List, Optional, Tuple, Union
from enum import Enum
from datetime import datetime
from pathlib import Path

logger = logging.getLogger(__name__)


class SourceType(Enum):
    """Knowledge source types."""
    WEB_SEARCH = "web_search"
    WEB_PAGE = "web_page"
    API = "api"
    PDF = "pdf"
    DOCX = "docx"
    TXT = "txt"
    MARKDOWN = "markdown"
    HTML = "html"
    IMAGE = "image"
    JSON = "json"
    CSV = "csv"
    EXCEL = "excel"
    CODE = "code"
    AUDIO = "audio"
    VIDEO = "video"
    CUSTOM = "custom"


@dataclass
class KnowledgeChunk:
    """A chunk of knowledge with metadata."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    content: str = ""
    source: str = ""
    source_type: SourceType = SourceType.TXT
    metadata: Dict[str, Any] = field(default_factory=dict)
    embedding: Optional[List[float]] = None
    timestamp: str = field(default_factory=lambda: datetime.now().isoformat())
    
    def to_dict(self) -> Dict[str, Any]:
        """Convert to dictionary for storage."""
        return {
            "id": self.id,
            "content": self.content,
            "source": self.source,
            "source_type": self.source_type.value,
            "metadata": self.metadata,
            "embedding": self.embedding,
            "timestamp": self.timestamp,
        }
    
    @classmethod
    def from_dict(cls, data: Dict[str, Any]) -> "KnowledgeChunk":
        return cls(
            id=data.get("id", str(uuid.uuid4())),
            content=data.get("content", ""),
            source=data.get("source", ""),
            source_type=SourceType(data.get("source_type", "txt")),
            metadata=data.get("metadata", {}),
            embedding=data.get("embedding"),
            timestamp=data.get("timestamp", datetime.now().isoformat()),
        )


@dataclass
class EmbeddingOutput:
    """Embedding output ready for vector database storage."""
    id: str
    embedding: List[float]
    content: str
    metadata: Dict[str, Any]
    
    def to_qdrant_point(self, collection_name: str = "default"):
        """Convert to Qdrant point format."""
        return {
            "id": self.id,
            "vector": self.embedding,
            "payload": {
                "content": self.content,
                **self.metadata,
            }
        }
    
    def to_pinecone_vector(self):
        """Convert to Pinecone vector format."""
        return {
            "id": self.id,
            "values": self.embedding,
            "metadata": {
                "content": self.content,
                **self.metadata,
            }
        }
    
    def to_weaviate_object(self, class_name: str = "Document"):
        """Convert to Weaviate object format."""
        return {
            "class": class_name,
            "id": self.id,
            "properties": {
                "content": self.content,
                **self.metadata,
            },
            "vector": self.embedding,
        }
    
    def to_chroma_document(self):
        """Convert to ChromaDB document format."""
        return {
            "ids": [self.id],
            "embeddings": [self.embedding],
            "documents": [self.content],
            "metadatas": [self.metadata],
        }
    
    def to_milvus_entity(self):
        """Convert to Milvus entity format."""
        return {
            "id": self.id,
            "vector": self.embedding,
            "content": self.content,
            **self.metadata,
        }
    
    def to_mongodb_document(self):
        """Convert to MongoDB Atlas Vector Search format."""
        return {
            "_id": self.id,
            "embedding": self.embedding,
            "content": self.content,
            "metadata": self.metadata,
        }
    
    def to_opensearch_document(self, index_name: str = "knowledge"):
        """Convert to OpenSearch format."""
        return {
            "_index": index_name,
            "_id": self.id,
            "_source": {
                "embedding": self.embedding,
                "content": self.content,
                **self.metadata,
            }
        }
    
    def to_pgvector_row(self):
        """Convert to pgvector format (PostgreSQL)."""
        return {
            "id": self.id,
            "embedding": self.embedding,  # Use pgvector extension
            "content": self.content,
            "metadata": json.dumps(self.metadata),
        }


class EmbeddingProvider(ABC):
    """Base class for embedding providers."""
    
    @abstractmethod
    def embed(self, text: str) -> List[float]:
        """Generate embedding for text."""
        pass
    
    @abstractmethod
    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        """Generate embeddings for multiple texts."""
        pass
    
    @property
    @abstractmethod
    def dimension(self) -> int:
        """Get embedding dimension."""
        pass


class OpenAIEmbedding(EmbeddingProvider):
    """OpenAI embedding provider."""
    
    def __init__(
        self,
        model: str = "text-embedding-3-small",
        api_key: Optional[str] = None,
    ):
        self.model = model
        self.api_key = api_key or os.getenv("OPENAI_API_KEY")
        self._client = None
        
        # Model dimensions
        self._dimensions = {
            "text-embedding-3-small": 1536,
            "text-embedding-3-large": 3072,
            "text-embedding-ada-002": 1536,
        }
    
    def _get_client(self):
        if not self._client:
            try:
                from openai import OpenAI
                self._client = OpenAI(api_key=self.api_key)
            except ImportError:
                raise ImportError("OpenAI embeddings require: pip install openai")
        return self._client
    
    def embed(self, text: str) -> List[float]:
        client = self._get_client()
        response = client.embeddings.create(input=text, model=self.model)
        return response.data[0].embedding
    
    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        client = self._get_client()
        response = client.embeddings.create(input=texts, model=self.model)
        return [item.embedding for item in response.data]
    
    @property
    def dimension(self) -> int:
        return self._dimensions.get(self.model, 1536)


class AzureOpenAIEmbedding(EmbeddingProvider):
    """Azure OpenAI embedding provider."""
    
    def __init__(
        self,
        deployment: str = "text-embedding-ada-002",
        api_key: Optional[str] = None,
        endpoint: Optional[str] = None,
        api_version: str = "2024-02-01",
    ):
        self.deployment = deployment
        self.api_key = api_key or os.getenv("AZURE_OPENAI_API_KEY")
        self.endpoint = endpoint or os.getenv("AZURE_OPENAI_ENDPOINT")
        self.api_version = api_version
        self._client = None
    
    def _get_client(self):
        if not self._client:
            try:
                from openai import AzureOpenAI
                self._client = AzureOpenAI(
                    api_key=self.api_key,
                    api_version=self.api_version,
                    azure_endpoint=self.endpoint,
                )
            except ImportError:
                raise ImportError("Azure OpenAI requires: pip install openai")
        return self._client
    
    def embed(self, text: str) -> List[float]:
        client = self._get_client()
        response = client.embeddings.create(input=text, model=self.deployment)
        return response.data[0].embedding
    
    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        client = self._get_client()
        response = client.embeddings.create(input=texts, model=self.deployment)
        return [item.embedding for item in response.data]
    
    @property
    def dimension(self) -> int:
        return 1536


class HuggingFaceEmbedding(EmbeddingProvider):
    """HuggingFace local embedding provider."""
    
    def __init__(
        self,
        model_name: str = "sentence-transformers/all-MiniLM-L6-v2",
        device: str = "cpu",
    ):
        self.model_name = model_name
        self.device = device
        self._model = None
        self._tokenizer = None
    
    def _load_model(self):
        if self._model is None:
            try:
                from sentence_transformers import SentenceTransformer
                self._model = SentenceTransformer(self.model_name, device=self.device)
            except ImportError:
                raise ImportError("HuggingFace embeddings require: pip install sentence-transformers")
    
    def embed(self, text: str) -> List[float]:
        self._load_model()
        return self._model.encode(text).tolist()
    
    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        self._load_model()
        return self._model.encode(texts).tolist()
    
    @property
    def dimension(self) -> int:
        self._load_model()
        return self._model.get_sentence_embedding_dimension()


class CohereEmbedding(EmbeddingProvider):
    """Cohere embedding provider."""
    
    def __init__(
        self,
        model: str = "embed-english-v3.0",
        api_key: Optional[str] = None,
        input_type: str = "search_document",
    ):
        self.model = model
        self.api_key = api_key or os.getenv("COHERE_API_KEY")
        self.input_type = input_type
        self._client = None
    
    def _get_client(self):
        if not self._client:
            try:
                import cohere
                self._client = cohere.Client(self.api_key)
            except ImportError:
                raise ImportError("Cohere embeddings require: pip install cohere")
        return self._client
    
    def embed(self, text: str) -> List[float]:
        client = self._get_client()
        response = client.embed(texts=[text], model=self.model, input_type=self.input_type)
        return response.embeddings[0]
    
    def embed_batch(self, texts: List[str]) -> List[List[float]]:
        client = self._get_client()
        response = client.embed(texts=texts, model=self.model, input_type=self.input_type)
        return response.embeddings
    
    @property
    def dimension(self) -> int:
        return 1024 if "v3" in self.model else 4096


class SourceLoader(ABC):
    """Base class for loading data from various sources."""
    
    @abstractmethod
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        """Load and chunk content from source."""
        pass
    
    @abstractmethod
    def supports(self, source: str) -> bool:
        """Check if loader supports the source."""
        pass


class TextLoader(SourceLoader):
    """Load text files."""
    
    def __init__(self, chunk_size: int = 1000, chunk_overlap: int = 200):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
    
    def supports(self, source: str) -> bool:
        ext = Path(source).suffix.lower()
        return ext in [".txt", ".text", ".log", ""]
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        with open(source, "r", encoding="utf-8") as f:
            content = f.read()
        
        return self._chunk_text(content, source, SourceType.TXT)
    
    def _chunk_text(
        self,
        text: str,
        source: str,
        source_type: SourceType,
    ) -> List[KnowledgeChunk]:
        """Split text into chunks."""
        chunks = []
        start = 0
        
        while start < len(text):
            end = start + self.chunk_size
            chunk_text = text[start:end]
            
            # Try to break at sentence or paragraph
            if end < len(text):
                for sep in ["\n\n", "\n", ". ", " "]:
                    last_sep = chunk_text.rfind(sep)
                    if last_sep > self.chunk_size // 2:
                        chunk_text = chunk_text[:last_sep + len(sep)]
                        break
            
            if chunk_text.strip():
                chunks.append(KnowledgeChunk(
                    content=chunk_text.strip(),
                    source=source,
                    source_type=source_type,
                    metadata={
                        "chunk_index": len(chunks),
                        "start_char": start,
                    }
                ))
            
            start += len(chunk_text) - self.chunk_overlap
            if start >= end:
                start = end
        
        return chunks


class MarkdownLoader(TextLoader):
    """Load Markdown files."""
    
    def supports(self, source: str) -> bool:
        ext = Path(source).suffix.lower()
        return ext in [".md", ".markdown"]
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        with open(source, "r", encoding="utf-8") as f:
            content = f.read()
        
        return self._chunk_text(content, source, SourceType.MARKDOWN)


class PDFLoader(SourceLoader):
    """Load PDF documents."""
    
    def __init__(self, chunk_size: int = 1000, chunk_overlap: int = 200):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
    
    def supports(self, source: str) -> bool:
        return Path(source).suffix.lower() == ".pdf"
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        try:
            import pypdf
        except ImportError:
            try:
                import PyPDF2 as pypdf
            except ImportError:
                raise ImportError("PDF loading requires: pip install pypdf")
        
        chunks = []
        
        with open(source, "rb") as f:
            reader = pypdf.PdfReader(f)
            
            for page_num, page in enumerate(reader.pages):
                text = page.extract_text()
                if text:
                    chunk = KnowledgeChunk(
                        content=text,
                        source=source,
                        source_type=SourceType.PDF,
                        metadata={
                            "page": page_num + 1,
                            "total_pages": len(reader.pages),
                        }
                    )
                    chunks.append(chunk)
        
        return chunks


class DocxLoader(SourceLoader):
    """Load Word documents."""
    
    def supports(self, source: str) -> bool:
        return Path(source).suffix.lower() in [".docx", ".doc"]
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        try:
            from docx import Document
        except ImportError:
            raise ImportError("DOCX loading requires: pip install python-docx")
        
        doc = Document(source)
        content = "\n".join([para.text for para in doc.paragraphs if para.text])
        
        return [KnowledgeChunk(
            content=content,
            source=source,
            source_type=SourceType.DOCX,
            metadata={"paragraphs": len(doc.paragraphs)}
        )]


class ImageLoader(SourceLoader):
    """Load and extract text from images using OCR or vision models."""
    
    def __init__(
        self,
        ocr_provider: str = "pytesseract",  # or "openai_vision", "azure_vision"
        api_key: Optional[str] = None,
    ):
        self.ocr_provider = ocr_provider
        self.api_key = api_key
    
    def supports(self, source: str) -> bool:
        ext = Path(source).suffix.lower()
        return ext in [".png", ".jpg", ".jpeg", ".gif", ".bmp", ".tiff", ".webp"]
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        if self.ocr_provider == "pytesseract":
            return self._load_with_tesseract(source)
        elif self.ocr_provider == "openai_vision":
            return self._load_with_openai_vision(source, **kwargs)
        else:
            return self._load_basic(source)
    
    def _load_basic(self, source: str) -> List[KnowledgeChunk]:
        """Return metadata only without OCR."""
        import base64
        
        with open(source, "rb") as f:
            data = f.read()
            b64 = base64.b64encode(data).decode()
        
        return [KnowledgeChunk(
            content=f"[Image: {Path(source).name}]",
            source=source,
            source_type=SourceType.IMAGE,
            metadata={
                "size_bytes": len(data),
                "base64_prefix": b64[:100] + "...",
            }
        )]
    
    def _load_with_tesseract(self, source: str) -> List[KnowledgeChunk]:
        """Extract text using Tesseract OCR."""
        try:
            import pytesseract
            from PIL import Image
        except ImportError:
            raise ImportError("OCR requires: pip install pytesseract pillow")
        
        image = Image.open(source)
        text = pytesseract.image_to_string(image)
        
        return [KnowledgeChunk(
            content=text,
            source=source,
            source_type=SourceType.IMAGE,
            metadata={
                "ocr_provider": "tesseract",
                "image_size": image.size,
            }
        )]
    
    def _load_with_openai_vision(
        self,
        source: str,
        prompt: str = "Extract all text and describe the content of this image.",
    ) -> List[KnowledgeChunk]:
        """Extract content using OpenAI Vision."""
        try:
            from openai import OpenAI
            import base64
        except ImportError:
            raise ImportError("OpenAI Vision requires: pip install openai")
        
        with open(source, "rb") as f:
            b64_image = base64.b64encode(f.read()).decode()
        
        client = OpenAI(api_key=self.api_key or os.getenv("OPENAI_API_KEY"))
        
        response = client.chat.completions.create(
            model="gpt-4o",
            messages=[{
                "role": "user",
                "content": [
                    {"type": "text", "text": prompt},
                    {
                        "type": "image_url",
                        "image_url": {"url": f"data:image/jpeg;base64,{b64_image}"}
                    }
                ]
            }]
        )
        
        content = response.choices[0].message.content
        
        return [KnowledgeChunk(
            content=content,
            source=source,
            source_type=SourceType.IMAGE,
            metadata={
                "ocr_provider": "openai_vision",
                "model": "gpt-4o",
            }
        )]


class WebLoader(SourceLoader):
    """Load content from web pages."""
    
    def __init__(self, chunk_size: int = 1000):
        self.chunk_size = chunk_size
    
    def supports(self, source: str) -> bool:
        return source.startswith(("http://", "https://"))
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        try:
            import requests
            from bs4 import BeautifulSoup
        except ImportError:
            raise ImportError("Web loading requires: pip install requests beautifulsoup4")
        
        response = requests.get(source, timeout=30)
        soup = BeautifulSoup(response.text, "html.parser")
        
        # Remove script and style elements
        for element in soup(["script", "style", "nav", "footer", "header"]):
            element.decompose()
        
        text = soup.get_text(separator="\n")
        # Clean up whitespace
        text = re.sub(r"\n\s*\n", "\n\n", text)
        text = text.strip()
        
        return [KnowledgeChunk(
            content=text,
            source=source,
            source_type=SourceType.WEB_PAGE,
            metadata={
                "title": soup.title.string if soup.title else "",
                "url": source,
            }
        )]


class WebSearchLoader(SourceLoader):
    """Load content from web search results."""
    
    def __init__(
        self,
        search_provider: str = "serper",  # or "google", "bing", "duckduckgo"
        api_key: Optional[str] = None,
        num_results: int = 5,
    ):
        self.search_provider = search_provider
        self.api_key = api_key
        self.num_results = num_results
        self._web_loader = WebLoader()
    
    def supports(self, source: str) -> bool:
        return source.startswith("search:")
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        query = source.replace("search:", "").strip()
        
        if self.search_provider == "serper":
            return self._search_serper(query)
        elif self.search_provider == "duckduckgo":
            return self._search_duckduckgo(query)
        else:
            logger.warning(f"Unknown search provider: {self.search_provider}")
            return []
    
    def _search_serper(self, query: str) -> List[KnowledgeChunk]:
        """Search using Serper API."""
        try:
            import requests
        except ImportError:
            raise ImportError("Web search requires: pip install requests")
        
        api_key = self.api_key or os.getenv("SERPER_API_KEY")
        
        response = requests.post(
            "https://google.serper.dev/search",
            headers={"X-API-KEY": api_key, "Content-Type": "application/json"},
            json={"q": query, "num": self.num_results}
        )
        
        data = response.json()
        chunks = []
        
        for result in data.get("organic", []):
            chunk = KnowledgeChunk(
                content=f"{result.get('title', '')}\n{result.get('snippet', '')}",
                source=result.get("link", ""),
                source_type=SourceType.WEB_SEARCH,
                metadata={
                    "title": result.get("title"),
                    "url": result.get("link"),
                    "query": query,
                }
            )
            chunks.append(chunk)
        
        return chunks
    
    def _search_duckduckgo(self, query: str) -> List[KnowledgeChunk]:
        """Search using DuckDuckGo."""
        try:
            from duckduckgo_search import DDGS
        except ImportError:
            raise ImportError("DuckDuckGo search requires: pip install duckduckgo-search")
        
        chunks = []
        
        with DDGS() as ddgs:
            for result in ddgs.text(query, max_results=self.num_results):
                chunk = KnowledgeChunk(
                    content=f"{result.get('title', '')}\n{result.get('body', '')}",
                    source=result.get("href", ""),
                    source_type=SourceType.WEB_SEARCH,
                    metadata={
                        "title": result.get("title"),
                        "url": result.get("href"),
                        "query": query,
                    }
                )
                chunks.append(chunk)
        
        return chunks


class APILoader(SourceLoader):
    """Load content from REST APIs."""
    
    def supports(self, source: str) -> bool:
        return source.startswith("api:")
    
    def load(
        self,
        source: str,
        method: str = "GET",
        headers: Optional[Dict] = None,
        data: Optional[Dict] = None,
        json_path: Optional[str] = None,
        **kwargs,
    ) -> List[KnowledgeChunk]:
        try:
            import requests
        except ImportError:
            raise ImportError("API loading requires: pip install requests")
        
        url = source.replace("api:", "").strip()
        
        response = requests.request(
            method=method,
            url=url,
            headers=headers or {},
            json=data,
            timeout=30,
        )
        
        try:
            content = response.json()
            
            # Extract specific path if provided
            if json_path:
                for key in json_path.split("."):
                    content = content.get(key, content)
            
            content_str = json.dumps(content, indent=2) if isinstance(content, (dict, list)) else str(content)
            
        except json.JSONDecodeError:
            content_str = response.text
        
        return [KnowledgeChunk(
            content=content_str,
            source=url,
            source_type=SourceType.API,
            metadata={
                "status_code": response.status_code,
                "method": method,
            }
        )]


class JSONLoader(SourceLoader):
    """Load JSON files."""
    
    def supports(self, source: str) -> bool:
        return Path(source).suffix.lower() == ".json"
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        with open(source, "r", encoding="utf-8") as f:
            data = json.load(f)
        
        return [KnowledgeChunk(
            content=json.dumps(data, indent=2),
            source=source,
            source_type=SourceType.JSON,
            metadata={"type": type(data).__name__}
        )]


class CSVLoader(SourceLoader):
    """Load CSV files."""
    
    def supports(self, source: str) -> bool:
        return Path(source).suffix.lower() == ".csv"
    
    def load(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        try:
            import pandas as pd
        except ImportError:
            # Fallback to csv module
            import csv
            chunks = []
            with open(source, "r", encoding="utf-8") as f:
                reader = csv.DictReader(f)
                for i, row in enumerate(reader):
                    chunks.append(KnowledgeChunk(
                        content=json.dumps(row),
                        source=source,
                        source_type=SourceType.CSV,
                        metadata={"row_index": i}
                    ))
            return chunks
        
        df = pd.read_csv(source)
        
        chunks = []
        for i, row in df.iterrows():
            chunks.append(KnowledgeChunk(
                content=row.to_json(),
                source=source,
                source_type=SourceType.CSV,
                metadata={
                    "row_index": i,
                    "columns": list(df.columns),
                }
            ))
        
        return chunks


class KnowledgeBuilder:
    """
    Build knowledge bases from various sources with embedding generation.
    
    Example:
        >>> from agenticaiframework.knowledge import KnowledgeBuilder
        >>> 
        >>> # Initialize with embedding provider
        >>> builder = KnowledgeBuilder(
        ...     embedding_provider="openai",
        ...     embedding_model="text-embedding-3-small"
        ... )
        >>> 
        >>> # Add knowledge from various sources
        >>> builder.add_from_file("docs/manual.pdf")
        >>> builder.add_from_url("https://example.com/article")
        >>> builder.add_from_web_search("machine learning best practices")
        >>> builder.add_from_api("https://api.example.com/data")
        >>> builder.add_from_image("diagram.png")
        >>> 
        >>> # Get embeddings for vector database
        >>> embeddings = builder.get_embeddings()
        >>> 
        >>> # Store in Qdrant
        >>> qdrant_points = [e.to_qdrant_point() for e in embeddings]
        >>> 
        >>> # Or store in Pinecone
        >>> pinecone_vectors = [e.to_pinecone_vector() for e in embeddings]
    """
    
    def __init__(
        self,
        embedding_provider: str = "openai",
        embedding_model: Optional[str] = None,
        chunk_size: int = 1000,
        chunk_overlap: int = 200,
    ):
        self.chunk_size = chunk_size
        self.chunk_overlap = chunk_overlap
        self.chunks: List[KnowledgeChunk] = []
        
        # Initialize embedding provider
        self.embedding = self._create_embedding_provider(
            embedding_provider,
            embedding_model,
        )
        
        # Initialize loaders
        self.loaders: List[SourceLoader] = [
            TextLoader(chunk_size, chunk_overlap),
            MarkdownLoader(chunk_size, chunk_overlap),
            PDFLoader(chunk_size, chunk_overlap),
            DocxLoader(),
            ImageLoader(),
            WebLoader(chunk_size),
            WebSearchLoader(),
            APILoader(),
            JSONLoader(),
            CSVLoader(),
        ]
    
    def _create_embedding_provider(
        self,
        provider: str,
        model: Optional[str],
    ) -> EmbeddingProvider:
        """Create embedding provider instance."""
        if provider == "openai":
            return OpenAIEmbedding(model=model or "text-embedding-3-small")
        elif provider == "azure":
            return AzureOpenAIEmbedding(deployment=model or "text-embedding-ada-002")
        elif provider == "huggingface":
            return HuggingFaceEmbedding(model_name=model or "sentence-transformers/all-MiniLM-L6-v2")
        elif provider == "cohere":
            return CohereEmbedding(model=model or "embed-english-v3.0")
        else:
            raise ValueError(f"Unknown embedding provider: {provider}")
    
    def add_loader(self, loader: SourceLoader) -> None:
        """Add a custom source loader."""
        self.loaders.insert(0, loader)
    
    def _find_loader(self, source: str) -> Optional[SourceLoader]:
        """Find appropriate loader for source."""
        for loader in self.loaders:
            if loader.supports(source):
                return loader
        return None
    
    def add(self, source: str, **kwargs) -> List[KnowledgeChunk]:
        """
        Add knowledge from any supported source.
        
        Args:
            source: File path, URL, or search query
            **kwargs: Additional arguments for the loader
            
        Returns:
            List of created knowledge chunks
        """
        loader = self._find_loader(source)
        if not loader:
            raise ValueError(f"No loader found for source: {source}")
        
        new_chunks = loader.load(source, **kwargs)
        self.chunks.extend(new_chunks)
        logger.info(f"Added {len(new_chunks)} chunks from: {source}")
        return new_chunks
    
    def add_from_file(self, file_path: str, **kwargs) -> List[KnowledgeChunk]:
        """Add knowledge from a file."""
        return self.add(file_path, **kwargs)
    
    def add_from_files(self, file_paths: List[str], **kwargs) -> List[KnowledgeChunk]:
        """Add knowledge from multiple files."""
        all_chunks = []
        for path in file_paths:
            all_chunks.extend(self.add_from_file(path, **kwargs))
        return all_chunks
    
    def add_from_directory(
        self,
        directory: str,
        extensions: Optional[List[str]] = None,
        recursive: bool = True,
    ) -> List[KnowledgeChunk]:
        """Add knowledge from all files in a directory."""
        extensions = extensions or [".txt", ".md", ".pdf", ".docx", ".json"]
        all_chunks = []
        
        path = Path(directory)
        pattern = "**/*" if recursive else "*"
        
        for file_path in path.glob(pattern):
            if file_path.is_file() and file_path.suffix.lower() in extensions:
                try:
                    chunks = self.add_from_file(str(file_path))
                    all_chunks.extend(chunks)
                except Exception as e:
                    logger.warning(f"Failed to load {file_path}: {e}")
        
        return all_chunks
    
    def add_from_url(self, url: str, **kwargs) -> List[KnowledgeChunk]:
        """Add knowledge from a web page."""
        return self.add(url, **kwargs)
    
    def add_from_urls(self, urls: List[str], **kwargs) -> List[KnowledgeChunk]:
        """Add knowledge from multiple URLs."""
        all_chunks = []
        for url in urls:
            try:
                chunks = self.add_from_url(url, **kwargs)
                all_chunks.extend(chunks)
            except Exception as e:
                logger.warning(f"Failed to load URL {url}: {e}")
        return all_chunks
    
    def add_from_web_search(
        self,
        query: str,
        num_results: int = 5,
        fetch_content: bool = False,
    ) -> List[KnowledgeChunk]:
        """
        Add knowledge from web search results.
        
        Args:
            query: Search query
            num_results: Number of results to fetch
            fetch_content: Whether to fetch full page content
        """
        # Update search loader results count
        for loader in self.loaders:
            if isinstance(loader, WebSearchLoader):
                loader.num_results = num_results
                break
        
        chunks = self.add(f"search:{query}")
        
        if fetch_content:
            # Fetch full content from search result URLs
            urls = [c.metadata.get("url") for c in chunks if c.metadata.get("url")]
            for url in urls:
                try:
                    self.add_from_url(url)
                except Exception as e:
                    logger.warning(f"Failed to fetch content from {url}: {e}")
        
        return chunks
    
    def add_from_api(
        self,
        url: str,
        method: str = "GET",
        headers: Optional[Dict] = None,
        data: Optional[Dict] = None,
        json_path: Optional[str] = None,
    ) -> List[KnowledgeChunk]:
        """Add knowledge from an API endpoint."""
        return self.add(
            f"api:{url}",
            method=method,
            headers=headers,
            data=data,
            json_path=json_path,
        )
    
    def add_from_image(
        self,
        image_path: str,
        ocr_provider: str = "pytesseract",
        vision_prompt: Optional[str] = None,
    ) -> List[KnowledgeChunk]:
        """
        Add knowledge from an image (OCR or vision model).
        
        Args:
            image_path: Path to image file
            ocr_provider: 'pytesseract' or 'openai_vision'
            vision_prompt: Custom prompt for vision model
        """
        # Update image loader
        for loader in self.loaders:
            if isinstance(loader, ImageLoader):
                loader.ocr_provider = ocr_provider
                break
        
        return self.add(image_path, prompt=vision_prompt)
    
    def add_text(self, text: str, source: str = "direct", metadata: Optional[Dict] = None) -> KnowledgeChunk:
        """Add text directly as knowledge."""
        chunk = KnowledgeChunk(
            content=text,
            source=source,
            source_type=SourceType.TXT,
            metadata=metadata or {},
        )
        self.chunks.append(chunk)
        return chunk
    
    def generate_embeddings(self, batch_size: int = 100) -> None:
        """Generate embeddings for all chunks."""
        chunks_without_embeddings = [c for c in self.chunks if c.embedding is None]
        
        for i in range(0, len(chunks_without_embeddings), batch_size):
            batch = chunks_without_embeddings[i:i + batch_size]
            texts = [c.content for c in batch]
            
            embeddings = self.embedding.embed_batch(texts)
            
            for chunk, embedding in zip(batch, embeddings):
                chunk.embedding = embedding
        
        logger.info(f"Generated embeddings for {len(chunks_without_embeddings)} chunks")
    
    def get_embeddings(self) -> List[EmbeddingOutput]:
        """
        Get all embeddings ready for vector database storage.
        
        Returns:
            List of EmbeddingOutput objects with vector DB conversion methods
        """
        # Generate embeddings if not already done
        self.generate_embeddings()
        
        outputs = []
        for chunk in self.chunks:
            if chunk.embedding:
                outputs.append(EmbeddingOutput(
                    id=chunk.id,
                    embedding=chunk.embedding,
                    content=chunk.content,
                    metadata={
                        "source": chunk.source,
                        "source_type": chunk.source_type.value,
                        "timestamp": chunk.timestamp,
                        **chunk.metadata,
                    }
                ))
        
        return outputs
    
    def get_chunks(self) -> List[KnowledgeChunk]:
        """Get all knowledge chunks."""
        return self.chunks
    
    def clear(self) -> None:
        """Clear all chunks."""
        self.chunks = []
    
    def export_to_json(self, file_path: str) -> None:
        """Export knowledge base to JSON file."""
        data = [chunk.to_dict() for chunk in self.chunks]
        with open(file_path, "w", encoding="utf-8") as f:
            json.dump(data, f, indent=2)
    
    def import_from_json(self, file_path: str) -> None:
        """Import knowledge base from JSON file."""
        with open(file_path, "r", encoding="utf-8") as f:
            data = json.load(f)
        
        for item in data:
            self.chunks.append(KnowledgeChunk.from_dict(item))


__all__ = [
    "SourceType",
    "KnowledgeChunk",
    "EmbeddingOutput",
    "EmbeddingProvider",
    "OpenAIEmbedding",
    "AzureOpenAIEmbedding",
    "HuggingFaceEmbedding",
    "CohereEmbedding",
    "SourceLoader",
    "TextLoader",
    "MarkdownLoader",
    "PDFLoader",
    "DocxLoader",
    "ImageLoader",
    "WebLoader",
    "WebSearchLoader",
    "APILoader",
    "JSONLoader",
    "CSVLoader",
    "KnowledgeBuilder",
]



"""
Enterprise Time Series Database Module.

Time series storage, aggregation, downsampling,
and analytics for metrics and IoT data.

Example:
    # Create time series database
    tsdb = create_timeseries_database()
    
    # Write data points
    await tsdb.write(
        metric="cpu_usage",
        value=75.5,
        tags={"host": "server1"},
    )
    
    # Query time range
    results = await tsdb.query(
        metric="cpu_usage",
        start=datetime.now() - timedelta(hours=1),
        end=datetime.now(),
        aggregation="avg",
        interval="5m",
    )
    
    # Downsampling
    await tsdb.downsample(
        metric="cpu_usage",
        from_interval="1m",
        to_interval="1h",
        aggregation="avg",
    )
"""

from __future__ import annotations

import asyncio
import bisect
import functools
import logging
import statistics
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class TimeSeriesError(Exception):
    """Time series error."""
    pass


class MetricNotFoundError(TimeSeriesError):
    """Metric not found."""
    pass


class Aggregation(str, Enum):
    """Aggregation functions."""
    AVG = "avg"
    SUM = "sum"
    MIN = "min"
    MAX = "max"
    COUNT = "count"
    FIRST = "first"
    LAST = "last"
    MEDIAN = "median"
    STDDEV = "stddev"
    PERCENTILE_50 = "p50"
    PERCENTILE_90 = "p90"
    PERCENTILE_95 = "p95"
    PERCENTILE_99 = "p99"


class RetentionPolicy(str, Enum):
    """Retention policies."""
    RAW = "raw"
    HOUR = "1h"
    DAY = "1d"
    WEEK = "1w"
    MONTH = "1mo"
    YEAR = "1y"


@dataclass
class DataPoint:
    """Single data point."""
    timestamp: datetime
    value: float
    tags: Dict[str, str] = field(default_factory=dict)
    
    def __lt__(self, other: "DataPoint") -> bool:
        return self.timestamp < other.timestamp


@dataclass
class Series:
    """Time series."""
    metric: str
    tags: Dict[str, str] = field(default_factory=dict)
    points: List[DataPoint] = field(default_factory=list)
    
    @property
    def key(self) -> str:
        """Get series key."""
        tag_str = ",".join(f"{k}={v}" for k, v in sorted(self.tags.items()))
        return f"{self.metric}{{{tag_str}}}" if tag_str else self.metric


@dataclass
class QueryResult:
    """Query result."""
    series: List[Series] = field(default_factory=list)
    execution_time_ms: float = 0.0
    
    def __iter__(self) -> Iterator[Series]:
        return iter(self.series)
    
    def __len__(self) -> int:
        return len(self.series)


@dataclass
class AggregatedPoint:
    """Aggregated data point."""
    timestamp: datetime
    value: float
    count: int = 1
    min_value: float = 0.0
    max_value: float = 0.0
    sum_value: float = 0.0


@dataclass
class MetricMetadata:
    """Metric metadata."""
    name: str
    description: str = ""
    unit: str = ""
    type: str = "gauge"  # gauge, counter, histogram
    tags: List[str] = field(default_factory=list)
    created_at: datetime = field(default_factory=datetime.utcnow)


@dataclass
class RetentionConfig:
    """Retention configuration."""
    raw_retention: timedelta = field(default_factory=lambda: timedelta(days=7))
    aggregation_intervals: List[Tuple[str, timedelta]] = field(
        default_factory=lambda: [
            ("1h", timedelta(days=30)),
            ("1d", timedelta(days=365)),
        ]
    )


@dataclass
class TimeSeriesStats:
    """Time series statistics."""
    total_metrics: int = 0
    total_series: int = 0
    total_points: int = 0
    oldest_point: Optional[datetime] = None
    newest_point: Optional[datetime] = None
    storage_bytes: int = 0


# Time series backend
class TimeSeriesBackend(ABC):
    """Abstract time series backend."""
    
    @abstractmethod
    async def write(
        self,
        metric: str,
        value: float,
        timestamp: datetime,
        tags: Dict[str, str],
    ) -> None:
        """Write data point."""
        pass
    
    @abstractmethod
    async def write_batch(
        self,
        points: List[Tuple[str, float, datetime, Dict[str, str]]],
    ) -> int:
        """Write batch of points."""
        pass
    
    @abstractmethod
    async def query(
        self,
        metric: str,
        start: datetime,
        end: datetime,
        tags: Optional[Dict[str, str]],
    ) -> List[DataPoint]:
        """Query data points."""
        pass
    
    @abstractmethod
    async def aggregate(
        self,
        metric: str,
        start: datetime,
        end: datetime,
        aggregation: Aggregation,
        interval: timedelta,
        tags: Optional[Dict[str, str]],
    ) -> List[AggregatedPoint]:
        """Aggregate data points."""
        pass
    
    @abstractmethod
    async def get_metrics(self) -> List[str]:
        """Get all metric names."""
        pass
    
    @abstractmethod
    async def get_stats(self) -> TimeSeriesStats:
        """Get statistics."""
        pass


class InMemoryTimeSeriesBackend(TimeSeriesBackend):
    """In-memory time series backend."""
    
    def __init__(self, max_points_per_series: int = 100000):
        self._data: Dict[str, Dict[str, List[DataPoint]]] = defaultdict(
            lambda: defaultdict(list)
        )
        self._metadata: Dict[str, MetricMetadata] = {}
        self._max_points = max_points_per_series
    
    def _get_series_key(self, tags: Dict[str, str]) -> str:
        """Get series key from tags."""
        return ",".join(f"{k}={v}" for k, v in sorted(tags.items())) or "_"
    
    async def write(
        self,
        metric: str,
        value: float,
        timestamp: datetime,
        tags: Dict[str, str],
    ) -> None:
        series_key = self._get_series_key(tags)
        points = self._data[metric][series_key]
        
        point = DataPoint(timestamp=timestamp, value=value, tags=tags)
        
        # Insert in sorted order
        bisect.insort(points, point)
        
        # Trim if needed
        if len(points) > self._max_points:
            self._data[metric][series_key] = points[-self._max_points:]
    
    async def write_batch(
        self,
        points: List[Tuple[str, float, datetime, Dict[str, str]]],
    ) -> int:
        count = 0
        for metric, value, timestamp, tags in points:
            await self.write(metric, value, timestamp, tags)
            count += 1
        return count
    
    async def query(
        self,
        metric: str,
        start: datetime,
        end: datetime,
        tags: Optional[Dict[str, str]],
    ) -> List[DataPoint]:
        results = []
        
        if metric not in self._data:
            return results
        
        for series_key, points in self._data[metric].items():
            for point in points:
                if start <= point.timestamp <= end:
                    if tags is None or all(
                        point.tags.get(k) == v for k, v in tags.items()
                    ):
                        results.append(point)
        
        return sorted(results, key=lambda p: p.timestamp)
    
    async def aggregate(
        self,
        metric: str,
        start: datetime,
        end: datetime,
        aggregation: Aggregation,
        interval: timedelta,
        tags: Optional[Dict[str, str]],
    ) -> List[AggregatedPoint]:
        points = await self.query(metric, start, end, tags)
        
        if not points:
            return []
        
        # Group by interval
        buckets: Dict[datetime, List[float]] = defaultdict(list)
        
        for point in points:
            # Round down to interval
            bucket_time = datetime.fromtimestamp(
                (point.timestamp.timestamp() // interval.total_seconds())
                * interval.total_seconds()
            )
            buckets[bucket_time].append(point.value)
        
        # Aggregate each bucket
        results = []
        
        for bucket_time, values in sorted(buckets.items()):
            agg_value = self._aggregate_values(values, aggregation)
            
            results.append(AggregatedPoint(
                timestamp=bucket_time,
                value=agg_value,
                count=len(values),
                min_value=min(values),
                max_value=max(values),
                sum_value=sum(values),
            ))
        
        return results
    
    def _aggregate_values(
        self,
        values: List[float],
        aggregation: Aggregation,
    ) -> float:
        """Apply aggregation to values."""
        if not values:
            return 0.0
        
        if aggregation == Aggregation.AVG:
            return statistics.mean(values)
        elif aggregation == Aggregation.SUM:
            return sum(values)
        elif aggregation == Aggregation.MIN:
            return min(values)
        elif aggregation == Aggregation.MAX:
            return max(values)
        elif aggregation == Aggregation.COUNT:
            return float(len(values))
        elif aggregation == Aggregation.FIRST:
            return values[0]
        elif aggregation == Aggregation.LAST:
            return values[-1]
        elif aggregation == Aggregation.MEDIAN:
            return statistics.median(values)
        elif aggregation == Aggregation.STDDEV:
            return statistics.stdev(values) if len(values) > 1 else 0.0
        elif aggregation == Aggregation.PERCENTILE_50:
            return self._percentile(values, 50)
        elif aggregation == Aggregation.PERCENTILE_90:
            return self._percentile(values, 90)
        elif aggregation == Aggregation.PERCENTILE_95:
            return self._percentile(values, 95)
        elif aggregation == Aggregation.PERCENTILE_99:
            return self._percentile(values, 99)
        
        return 0.0
    
    def _percentile(self, values: List[float], p: int) -> float:
        """Calculate percentile."""
        sorted_values = sorted(values)
        k = (len(sorted_values) - 1) * p / 100
        f = int(k)
        c = f + 1 if f < len(sorted_values) - 1 else f
        
        return sorted_values[f] + (k - f) * (sorted_values[c] - sorted_values[f])
    
    async def get_metrics(self) -> List[str]:
        return list(self._data.keys())
    
    async def get_stats(self) -> TimeSeriesStats:
        total_points = 0
        oldest = None
        newest = None
        
        for metric_data in self._data.values():
            for points in metric_data.values():
                total_points += len(points)
                
                if points:
                    if oldest is None or points[0].timestamp < oldest:
                        oldest = points[0].timestamp
                    if newest is None or points[-1].timestamp > newest:
                        newest = points[-1].timestamp
        
        return TimeSeriesStats(
            total_metrics=len(self._data),
            total_series=sum(
                len(series) for series in self._data.values()
            ),
            total_points=total_points,
            oldest_point=oldest,
            newest_point=newest,
            storage_bytes=total_points * 24,  # Estimate
        )
    
    async def delete_before(self, metric: str, before: datetime) -> int:
        """Delete points before timestamp."""
        deleted = 0
        
        if metric in self._data:
            for series_key, points in self._data[metric].items():
                original_count = len(points)
                self._data[metric][series_key] = [
                    p for p in points if p.timestamp >= before
                ]
                deleted += original_count - len(self._data[metric][series_key])
        
        return deleted


# Time series database
class TimeSeriesDatabase:
    """
    Time series database service.
    """
    
    def __init__(
        self,
        backend: Optional[TimeSeriesBackend] = None,
        retention: Optional[RetentionConfig] = None,
    ):
        self._backend = backend or InMemoryTimeSeriesBackend()
        self._retention = retention or RetentionConfig()
        self._metadata: Dict[str, MetricMetadata] = {}
    
    async def write(
        self,
        metric: str,
        value: float,
        timestamp: Optional[datetime] = None,
        **tags,
    ) -> None:
        """
        Write a data point.
        
        Args:
            metric: Metric name
            value: Value
            timestamp: Timestamp (default: now)
            **tags: Tags
        """
        ts = timestamp or datetime.utcnow()
        await self._backend.write(metric, value, ts, tags)
    
    async def write_batch(
        self,
        metric: str,
        points: List[Tuple[datetime, float]],
        **tags,
    ) -> int:
        """
        Write batch of points.
        
        Args:
            metric: Metric name
            points: List of (timestamp, value) tuples
            **tags: Tags for all points
            
        Returns:
            Number of points written
        """
        batch = [(metric, value, ts, tags) for ts, value in points]
        return await self._backend.write_batch(batch)
    
    async def query(
        self,
        metric: str,
        start: datetime,
        end: Optional[datetime] = None,
        aggregation: Optional[Aggregation] = None,
        interval: Optional[str] = None,
        **tags,
    ) -> QueryResult:
        """
        Query time series data.
        
        Args:
            metric: Metric name
            start: Start time
            end: End time (default: now)
            aggregation: Aggregation function
            interval: Aggregation interval (e.g., "5m", "1h")
            **tags: Tag filters
            
        Returns:
            Query result
        """
        import time
        start_time = time.perf_counter()
        
        end_time = end or datetime.utcnow()
        tag_filter = tags if tags else None
        
        if aggregation and interval:
            interval_td = self._parse_interval(interval)
            agg_points = await self._backend.aggregate(
                metric, start, end_time, aggregation, interval_td, tag_filter
            )
            
            series = Series(
                metric=metric,
                tags=tags,
                points=[
                    DataPoint(timestamp=ap.timestamp, value=ap.value)
                    for ap in agg_points
                ],
            )
            result_series = [series]
        else:
            points = await self._backend.query(metric, start, end_time, tag_filter)
            
            # Group by tags
            series_map: Dict[str, Series] = {}
            
            for point in points:
                key = ",".join(f"{k}={v}" for k, v in sorted(point.tags.items()))
                
                if key not in series_map:
                    series_map[key] = Series(metric=metric, tags=point.tags)
                
                series_map[key].points.append(point)
            
            result_series = list(series_map.values())
        
        return QueryResult(
            series=result_series,
            execution_time_ms=(time.perf_counter() - start_time) * 1000,
        )
    
    def _parse_interval(self, interval: str) -> timedelta:
        """Parse interval string."""
        unit = interval[-1]
        value = int(interval[:-1])
        
        if unit == "s":
            return timedelta(seconds=value)
        elif unit == "m":
            return timedelta(minutes=value)
        elif unit == "h":
            return timedelta(hours=value)
        elif unit == "d":
            return timedelta(days=value)
        elif unit == "w":
            return timedelta(weeks=value)
        
        raise ValueError(f"Invalid interval: {interval}")
    
    async def downsample(
        self,
        metric: str,
        from_interval: str,
        to_interval: str,
        aggregation: Aggregation = Aggregation.AVG,
        start: Optional[datetime] = None,
        end: Optional[datetime] = None,
    ) -> int:
        """
        Downsample data.
        
        Args:
            metric: Metric name
            from_interval: Source interval
            to_interval: Target interval
            aggregation: Aggregation function
            start: Start time
            end: End time
            
        Returns:
            Number of points created
        """
        end_time = end or datetime.utcnow()
        start_time = start or (end_time - timedelta(days=30))
        
        to_interval_td = self._parse_interval(to_interval)
        
        agg_points = await self._backend.aggregate(
            metric, start_time, end_time, aggregation, to_interval_td, None
        )
        
        # Write aggregated points with different metric name
        downsampled_metric = f"{metric}:{to_interval}"
        
        count = 0
        for point in agg_points:
            await self._backend.write(
                downsampled_metric,
                point.value,
                point.timestamp,
                {},
            )
            count += 1
        
        return count
    
    async def get_metrics(self) -> List[str]:
        """Get all metric names."""
        return await self._backend.get_metrics()
    
    async def get_stats(self) -> TimeSeriesStats:
        """Get database statistics."""
        return await self._backend.get_stats()
    
    async def register_metric(
        self,
        name: str,
        description: str = "",
        unit: str = "",
        metric_type: str = "gauge",
        tags: Optional[List[str]] = None,
    ) -> MetricMetadata:
        """
        Register metric metadata.
        
        Args:
            name: Metric name
            description: Description
            unit: Unit of measurement
            metric_type: Type (gauge, counter, histogram)
            tags: Expected tag keys
            
        Returns:
            Metric metadata
        """
        metadata = MetricMetadata(
            name=name,
            description=description,
            unit=unit,
            type=metric_type,
            tags=tags or [],
        )
        
        self._metadata[name] = metadata
        return metadata
    
    async def get_metric_metadata(
        self,
        name: str,
    ) -> Optional[MetricMetadata]:
        """Get metric metadata."""
        return self._metadata.get(name)
    
    async def apply_retention(self) -> Dict[str, int]:
        """
        Apply retention policy.
        
        Returns:
            Dict of metric -> deleted points count
        """
        if not isinstance(self._backend, InMemoryTimeSeriesBackend):
            return {}
        
        deleted: Dict[str, int] = {}
        cutoff = datetime.utcnow() - self._retention.raw_retention
        
        for metric in await self._backend.get_metrics():
            count = await self._backend.delete_before(metric, cutoff)
            if count > 0:
                deleted[metric] = count
        
        return deleted


# Helper classes
class MetricRecorder:
    """
    Helper for recording metrics.
    
    Example:
        recorder = MetricRecorder(tsdb, "api_requests")
        recorder.increment()
        recorder.record(1.5)
    """
    
    def __init__(
        self,
        tsdb: TimeSeriesDatabase,
        metric: str,
        **default_tags,
    ):
        self._tsdb = tsdb
        self._metric = metric
        self._default_tags = default_tags
        self._value = 0.0
    
    async def record(self, value: float, **extra_tags) -> None:
        """Record value."""
        tags = {**self._default_tags, **extra_tags}
        await self._tsdb.write(self._metric, value, **tags)
    
    async def increment(self, delta: float = 1.0, **extra_tags) -> None:
        """Increment counter."""
        self._value += delta
        await self.record(self._value, **extra_tags)
    
    async def gauge(self, value: float, **extra_tags) -> None:
        """Set gauge value."""
        self._value = value
        await self.record(value, **extra_tags)


# Decorators
def record_timing(
    tsdb: TimeSeriesDatabase,
    metric: str,
    **tags,
):
    """
    Decorator to record function timing.
    
    Args:
        tsdb: Time series database
        metric: Metric name
        **tags: Tags
    """
    def decorator(func):
        @functools.wraps(func)
        async def async_wrapper(*args, **kwargs):
            import time
            start = time.perf_counter()
            try:
                return await func(*args, **kwargs)
            finally:
                duration = (time.perf_counter() - start) * 1000
                await tsdb.write(metric, duration, **tags)
        
        @functools.wraps(func)
        def sync_wrapper(*args, **kwargs):
            import time
            start = time.perf_counter()
            try:
                return func(*args, **kwargs)
            finally:
                duration = (time.perf_counter() - start) * 1000
                asyncio.create_task(tsdb.write(metric, duration, **tags))
        
        if asyncio.iscoroutinefunction(func):
            return async_wrapper
        return sync_wrapper
    
    return decorator


# Factory functions
def create_timeseries_database(
    backend: Optional[TimeSeriesBackend] = None,
    retention: Optional[RetentionConfig] = None,
) -> TimeSeriesDatabase:
    """Create time series database."""
    return TimeSeriesDatabase(backend=backend, retention=retention)


def create_in_memory_backend(
    max_points: int = 100000,
) -> InMemoryTimeSeriesBackend:
    """Create in-memory backend."""
    return InMemoryTimeSeriesBackend(max_points_per_series=max_points)


def create_metric_recorder(
    tsdb: TimeSeriesDatabase,
    metric: str,
    **tags,
) -> MetricRecorder:
    """Create metric recorder."""
    return MetricRecorder(tsdb, metric, **tags)


__all__ = [
    # Exceptions
    "TimeSeriesError",
    "MetricNotFoundError",
    # Enums
    "Aggregation",
    "RetentionPolicy",
    # Data classes
    "DataPoint",
    "Series",
    "QueryResult",
    "AggregatedPoint",
    "MetricMetadata",
    "RetentionConfig",
    "TimeSeriesStats",
    # Backend
    "TimeSeriesBackend",
    "InMemoryTimeSeriesBackend",
    # Main class
    "TimeSeriesDatabase",
    # Helpers
    "MetricRecorder",
    # Decorators
    "record_timing",
    # Factory functions
    "create_timeseries_database",
    "create_in_memory_backend",
    "create_metric_recorder",
]



"""
Code Interpreter and Execution Tools.
"""

import logging
import sys
import io
import traceback
import shutil
import subprocess
import tempfile
import os
from typing import Any, Dict, List, Optional

from ..base import BaseTool, ToolConfig

logger = logging.getLogger(__name__)


class CodeInterpreterTool(BaseTool):
    """
    Tool for executing Python code safely.
    
    Features:
    - Python code execution
    - Sandboxed environment
    - Output capture
    - Variable persistence
    - Package installation
    """
    
    def __init__(
        self,
        config: Optional[ToolConfig] = None,
        allowed_modules: Optional[List[str]] = None,
        timeout: float = 30.0,
        max_output_length: int = 10000,
        allow_package_install: bool = True,
        install_timeout: float = 60.0,
    ):
        super().__init__(config or ToolConfig(
            name="CodeInterpreterTool",
            description="Execute Python code safely"
        ))
        self.allowed_modules = allowed_modules or [
            'math', 'json', 're', 'datetime', 'collections',
            'itertools', 'functools', 'random', 'statistics',
            'numpy', 'pandas', 'matplotlib',
        ]
        self.timeout = timeout
        self.max_output_length = max_output_length
        self.allow_package_install = allow_package_install
        self.install_timeout = install_timeout
        self._globals: Dict[str, Any] = {}
        self._locals: Dict[str, Any] = {}
    
    def _execute(
        self,
        code: str,
        capture_output: bool = True,
        persist_variables: bool = True,
        reset_environment: bool = False,
        packages: Optional[List[str]] = None,
        auto_install: bool = True,
        cleanup: bool = False,
        isolate_packages: bool = True,
    ) -> Dict[str, Any]:
        """
        Execute Python code.
        
        Args:
            code: Python code to execute
            capture_output: Capture stdout/stderr
            persist_variables: Keep variables between calls
            reset_environment: Reset execution environment
            
        Returns:
            Dict with execution results
        """
        if reset_environment:
            self._globals = {}
            self._locals = {}

        installed_packages = []
        temp_package_dir = None
        original_sys_path = None
        original_allowed = list(self.allowed_modules)
        if packages and auto_install:
            if not self.allow_package_install:
                return {
                    'code': code,
                    'status': 'error',
                    'output': None,
                    'error': 'Package installation is disabled',
                    'return_value': None,
                    'variables': {},
                }
            install_result = self._install_packages(packages, isolate=isolate_packages)
            if install_result.get('status') != 'success':
                return {
                    'code': code,
                    'status': 'error',
                    'output': None,
                    'error': install_result.get('error', 'Package installation failed'),
                    'return_value': None,
                    'variables': {},
                    'packages': install_result,
                }
            installed_packages = install_result.get('installed', [])
            temp_package_dir = install_result.get('install_dir')
            if temp_package_dir:
                original_sys_path = list(sys.path)
                sys.path.insert(0, temp_package_dir)
            # Temporarily allow installed packages
            self.allowed_modules = list(set(self.allowed_modules + [p.split('==')[0] for p in installed_packages]))
        
        # Set up safe globals
        safe_globals = {
            '__builtins__': self._get_safe_builtins(),
            **self._globals,
        }
        
        # Capture output
        old_stdout = sys.stdout
        old_stderr = sys.stderr
        
        stdout_capture = io.StringIO()
        stderr_capture = io.StringIO()
        
        if capture_output:
            sys.stdout = stdout_capture
            sys.stderr = stderr_capture
        
        result = {
            'code': code,
            'status': 'success',
            'output': None,
            'error': None,
            'return_value': None,
            'variables': {},
            'packages': installed_packages,
        }
        
        try:
            # Compile and execute
            compiled = compile(code, '<code>', 'exec')
            
            # Execute with timeout (simplified - real timeout needs threading)
            exec_locals = dict(self._locals) if persist_variables else {}
            exec(compiled, safe_globals, exec_locals)
            
            # Get last expression value if any
            try:
                last_expr = compile(code.split('\n')[-1], '<expr>', 'eval')
                result['return_value'] = eval(last_expr, safe_globals, exec_locals)
            except:
                pass
            
            # Update persistent state
            if persist_variables:
                self._locals.update(exec_locals)
                self._globals.update(
                    {k: v for k, v in safe_globals.items() if k != '__builtins__'}
                )
            
            # Get variable info
            result['variables'] = {
                k: type(v).__name__
                for k, v in exec_locals.items()
                if not k.startswith('_')
            }
            
        except Exception as e:
            result['status'] = 'error'
            result['error'] = {
                'type': type(e).__name__,
                'message': str(e),
                'traceback': traceback.format_exc(),
            }
        
        finally:
            sys.stdout = old_stdout
            sys.stderr = old_stderr
            if original_sys_path is not None:
                sys.path = original_sys_path
            self.allowed_modules = original_allowed
        
        if capture_output:
            stdout_val = stdout_capture.getvalue()
            stderr_val = stderr_capture.getvalue()
            
            result['output'] = stdout_val[:self.max_output_length]
            if stderr_val:
                result['stderr'] = stderr_val[:self.max_output_length]
            
            if len(stdout_val) > self.max_output_length:
                result['output_truncated'] = True
        
        if cleanup:
            self.reset()
            if temp_package_dir and os.path.isdir(temp_package_dir):
                try:
                    shutil.rmtree(temp_package_dir, ignore_errors=True)
                except Exception:
                    pass

        return result
    
    def _get_safe_builtins(self) -> Dict[str, Any]:
        """Get safe subset of builtins."""
        safe = {}
        
        # Safe types
        safe_names = [
            'True', 'False', 'None',
            'abs', 'all', 'any', 'bin', 'bool', 'bytearray', 'bytes',
            'callable', 'chr', 'complex', 'dict', 'divmod', 'enumerate',
            'filter', 'float', 'format', 'frozenset', 'hash', 'hex',
            'int', 'isinstance', 'issubclass', 'iter', 'len', 'list',
            'map', 'max', 'min', 'next', 'object', 'oct', 'ord', 'pow',
            'print', 'range', 'repr', 'reversed', 'round', 'set',
            'slice', 'sorted', 'str', 'sum', 'tuple', 'type', 'zip',
        ]
        
        import builtins
        for name in safe_names:
            if hasattr(builtins, name):
                safe[name] = getattr(builtins, name)
        
        # Safe import
        safe['__import__'] = self._safe_import
        
        return safe
    
    def _safe_import(
        self,
        name: str,
        globals: Dict = None,
        locals: Dict = None,
        fromlist: List = None,
        level: int = 0,
    ):
        """Safe import function."""
        # Check if module is allowed
        base_module = name.split('.')[0]
        
        if base_module not in self.allowed_modules:
            raise ImportError(
                f"Module '{name}' is not allowed. "
                f"Allowed modules: {self.allowed_modules}"
            )
        
        return __import__(name, globals, locals, fromlist or [], level)
    
    def get_variables(self) -> Dict[str, Any]:
        """Get current variable values."""
        return {
            k: repr(v)[:100]
            for k, v in self._locals.items()
            if not k.startswith('_')
        }
    
    def set_variable(self, name: str, value: Any):
        """Set a variable in the execution environment."""
        self._locals[name] = value
    
    def reset(self):
        """Reset the execution environment."""
        self._globals = {}
        self._locals = {}
    
    def install_package(self, package: str) -> Dict[str, Any]:
        """Install a Python package (use with caution)."""
        try:
            result = subprocess.run(
                [sys.executable, '-m', 'pip', 'install', package],
                capture_output=True,
                text=True,
                timeout=60,
            )
            
            return {
                'status': 'success' if result.returncode == 0 else 'error',
                'package': package,
                'stdout': result.stdout,
                'stderr': result.stderr,
            }
        except subprocess.TimeoutExpired:
            return {
                'status': 'error',
                'error': 'Installation timed out',
            }
        except Exception as e:
            return {
                'status': 'error',
                'error': str(e),
            }

    def _install_packages(self, packages: List[str], isolate: bool = True) -> Dict[str, Any]:
        """Install multiple Python packages."""
        installed = []
        errors = []
        install_dir = None

        if isolate:
            install_dir = tempfile.mkdtemp(prefix='agenticai_packages_')

        for package in packages:
            try:
                result = subprocess.run(
                    [
                        sys.executable,
                        '-m',
                        'pip',
                        'install',
                        package,
                        *(['--target', install_dir] if install_dir else []),
                    ],
                    capture_output=True,
                    text=True,
                    timeout=self.install_timeout,
                )
                if result.returncode == 0:
                    installed.append(package)
                else:
                    errors.append({
                        'package': package,
                        'stderr': result.stderr.strip(),
                        'stdout': result.stdout.strip(),
                    })
            except subprocess.TimeoutExpired:
                errors.append({
                    'package': package,
                    'error': 'Installation timed out',
                })
            except Exception as e:  # noqa: BLE001
                errors.append({
                    'package': package,
                    'error': str(e),
                })

        if errors:
            if install_dir and os.path.isdir(install_dir):
                shutil.rmtree(install_dir, ignore_errors=True)
            return {
                'status': 'error',
                'installed': installed,
                'errors': errors,
            }

        return {
            'status': 'success',
            'installed': installed,
            'install_dir': install_dir,
        }


class JavaScriptCodeInterpreterTool(BaseTool):
    """
    Tool for executing JavaScript code safely via Node.js.
    
    Features:
    - JavaScript execution using Node.js
    - Output capture
    - Timeout enforcement
    """

    def __init__(
        self,
        config: Optional[ToolConfig] = None,
        timeout: float = 10.0,
        max_output_length: int = 10000,
        node_path: Optional[str] = None,
        npm_path: Optional[str] = None,
        allow_package_install: bool = True,
        install_timeout: float = 120.0,
    ):
        super().__init__(config or ToolConfig(
            name="JavaScriptCodeInterpreterTool",
            description="Execute JavaScript code using Node.js"
        ))
        self.timeout = timeout
        self.max_output_length = max_output_length
        self.node_path = node_path or shutil.which('node')
        self.npm_path = npm_path or shutil.which('npm')
        self.allow_package_install = allow_package_install
        self.install_timeout = install_timeout

    def _execute(
        self,
        code: str,
        capture_output: bool = True,
        packages: Optional[List[str]] = None,
        auto_install: bool = True,
        cleanup: bool = False,
    ) -> Dict[str, Any]:
        """
        Execute JavaScript code via Node.js.

        Args:
            code: JavaScript code to execute
            capture_output: Capture stdout/stderr

        Returns:
            Dict with execution results
        """
        result = {
            'code': code,
            'status': 'success',
            'output': None,
            'error': None,
        }

        if not self.node_path:
            return {
                **result,
                'status': 'error',
                'error': 'Node.js not available on PATH',
            }

        work_dir = None
        try:
            use_temp_dir = bool(packages) or cleanup
            if use_temp_dir:
                work_dir = tempfile.TemporaryDirectory()

            if packages and auto_install:
                if not self.allow_package_install:
                    return {
                        **result,
                        'status': 'error',
                        'error': 'Package installation is disabled',
                    }
                if not self.npm_path:
                    return {
                        **result,
                        'status': 'error',
                        'error': 'npm not available on PATH',
                    }

                cwd = work_dir.name if work_dir else None
                init_cmd = [self.npm_path, 'init', '-y']
                subprocess.run(init_cmd, capture_output=True, text=True, timeout=self.install_timeout, cwd=cwd)

                install_cmd = [self.npm_path, 'install', *packages]
                install_result = subprocess.run(
                    install_cmd,
                    capture_output=True,
                    text=True,
                    timeout=self.install_timeout,
                    cwd=cwd,
                )
                if install_result.returncode != 0:
                    return {
                        **result,
                        'status': 'error',
                        'error': install_result.stderr.strip() or 'npm install failed',
                        'stderr': install_result.stderr[:self.max_output_length],
                    }

            with tempfile.NamedTemporaryFile('w', suffix='.js', delete=False, dir=work_dir.name if work_dir else None) as tmp:
                tmp.write(code)
                tmp_path = tmp.name

            completed = subprocess.run(
                [self.node_path, tmp_path],
                capture_output=capture_output,
                text=True,
                timeout=self.timeout,
                cwd=work_dir.name if work_dir else None,
            )

            stdout_val = completed.stdout or ""
            stderr_val = completed.stderr or ""

            result['output'] = stdout_val[:self.max_output_length]
            if stderr_val:
                result['stderr'] = stderr_val[:self.max_output_length]

            if completed.returncode != 0:
                result['status'] = 'error'
                result['error'] = stderr_val.strip() or f"Process exited with code {completed.returncode}"

            if len(stdout_val) > self.max_output_length:
                result['output_truncated'] = True

        except subprocess.TimeoutExpired:
            result['status'] = 'error'
            result['error'] = 'Execution timed out'
        except Exception as e:
            result['status'] = 'error'
            result['error'] = str(e)
        finally:
            try:
                if 'tmp_path' in locals():
                    import os
                    os.unlink(tmp_path)
            except Exception:
                pass
            if work_dir is not None:
                work_dir.cleanup()

        return result


__all__ = ['CodeInterpreterTool', 'JavaScriptCodeInterpreterTool']



"""
Enterprise Request Router Module.

Smart request routing, circuit breaking, retry policies,
and traffic management.

Example:
    # Create request router
    router = create_request_router()
    
    # Add route
    await router.add_route(
        pattern="/api/users/*",
        target="user-service",
        weight=100,
    )
    
    # Route request
    target = await router.route("/api/users/123")
    
    # With circuit breaker
    async with router.protected_call("user-service") as call:
        result = await call()
"""

from __future__ import annotations

import asyncio
import fnmatch
import logging
import random
import time
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, timedelta
from enum import Enum
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Pattern,
    Tuple,
)

logger = logging.getLogger(__name__)


class RoutingError(Exception):
    """Routing error."""
    pass


class NoRouteFoundError(RoutingError):
    """No route found error."""
    pass


class CircuitOpenError(RoutingError):
    """Circuit breaker open error."""
    pass


class RouteType(str, Enum):
    """Route type."""
    EXACT = "exact"
    PREFIX = "prefix"
    PATTERN = "pattern"
    REGEX = "regex"


class CircuitState(str, Enum):
    """Circuit breaker state."""
    CLOSED = "closed"
    OPEN = "open"
    HALF_OPEN = "half_open"


class RetryStrategy(str, Enum):
    """Retry strategy."""
    FIXED = "fixed"
    LINEAR = "linear"
    EXPONENTIAL = "exponential"
    FIBONACCI = "fibonacci"


class LoadBalanceStrategy(str, Enum):
    """Load balance strategy for multi-target routes."""
    ROUND_ROBIN = "round_robin"
    RANDOM = "random"
    WEIGHTED = "weighted"
    LEAST_LATENCY = "least_latency"


@dataclass
class RouteTarget:
    """Route target."""
    service: str = ""
    weight: int = 100
    
    # Metadata
    version: str = ""
    zone: str = ""
    
    # Statistics
    request_count: int = 0
    error_count: int = 0
    total_latency_ms: float = 0.0
    
    @property
    def avg_latency_ms(self) -> float:
        if self.request_count == 0:
            return 0.0
        return self.total_latency_ms / self.request_count
    
    @property
    def error_rate(self) -> float:
        if self.request_count == 0:
            return 0.0
        return self.error_count / self.request_count


@dataclass
class Route:
    """Route definition."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    
    # Pattern
    pattern: str = ""
    route_type: RouteType = RouteType.PREFIX
    
    # Targets
    targets: List[RouteTarget] = field(default_factory=list)
    lb_strategy: LoadBalanceStrategy = LoadBalanceStrategy.ROUND_ROBIN
    
    # Conditions
    methods: List[str] = field(default_factory=list)  # GET, POST, etc.
    headers: Dict[str, str] = field(default_factory=dict)
    
    # Priority
    priority: int = 0
    
    # Options
    strip_prefix: bool = False
    add_prefix: str = ""
    timeout_seconds: float = 30.0
    
    # Retry
    retry_enabled: bool = True
    max_retries: int = 3
    retry_strategy: RetryStrategy = RetryStrategy.EXPONENTIAL
    
    # Circuit breaker
    circuit_breaker_enabled: bool = True
    
    # State
    enabled: bool = True
    _rr_index: int = 0
    
    # Metadata
    name: str = ""
    description: str = ""
    tags: List[str] = field(default_factory=list)


@dataclass
class CircuitBreaker:
    """Circuit breaker."""
    name: str = ""
    
    # State
    state: CircuitState = CircuitState.CLOSED
    
    # Thresholds
    failure_threshold: int = 5
    success_threshold: int = 3
    
    # Timeouts
    timeout_seconds: float = 30.0
    
    # Counters
    failure_count: int = 0
    success_count: int = 0
    
    # Timing
    last_failure: Optional[datetime] = None
    opened_at: Optional[datetime] = None
    
    @property
    def is_open(self) -> bool:
        return self.state == CircuitState.OPEN
    
    @property
    def is_half_open(self) -> bool:
        return self.state == CircuitState.HALF_OPEN


@dataclass
class RetryPolicy:
    """Retry policy."""
    max_retries: int = 3
    strategy: RetryStrategy = RetryStrategy.EXPONENTIAL
    base_delay_seconds: float = 0.1
    max_delay_seconds: float = 10.0
    
    # Retry conditions
    retry_on_status_codes: List[int] = field(default_factory=lambda: [502, 503, 504])
    retry_on_exceptions: List[str] = field(default_factory=list)


@dataclass
class RoutingContext:
    """Routing context."""
    path: str = ""
    method: str = "GET"
    headers: Dict[str, str] = field(default_factory=dict)
    query_params: Dict[str, str] = field(default_factory=dict)
    
    # Client info
    client_ip: str = ""
    client_id: str = ""
    
    # Metadata
    trace_id: str = ""
    span_id: str = ""


@dataclass
class RoutingResult:
    """Routing result."""
    route: Route
    target: RouteTarget
    
    # Transformed path
    transformed_path: str = ""
    
    # Timing
    routing_time_ms: float = 0.0


@dataclass
class RouterStats:
    """Router statistics."""
    total_routes: int = 0
    active_routes: int = 0
    
    total_requests: int = 0
    successful_requests: int = 0
    failed_requests: int = 0
    
    avg_routing_time_ms: float = 0.0
    
    circuits_open: int = 0


# Route matcher
class RouteMatcher(ABC):
    """Route matcher."""
    
    @abstractmethod
    def matches(self, route: Route, context: RoutingContext) -> bool:
        pass


class DefaultRouteMatcher(RouteMatcher):
    """Default route matcher."""
    
    def matches(self, route: Route, context: RoutingContext) -> bool:
        # Check if route is enabled
        if not route.enabled:
            return False
        
        # Check method
        if route.methods and context.method.upper() not in route.methods:
            return False
        
        # Check headers
        for key, value in route.headers.items():
            if context.headers.get(key) != value:
                return False
        
        # Check pattern
        return self._match_pattern(route, context.path)
    
    def _match_pattern(self, route: Route, path: str) -> bool:
        if route.route_type == RouteType.EXACT:
            return path == route.pattern
        
        elif route.route_type == RouteType.PREFIX:
            return path.startswith(route.pattern)
        
        elif route.route_type == RouteType.PATTERN:
            return fnmatch.fnmatch(path, route.pattern)
        
        return False


# Target selector
class TargetSelector(ABC):
    """Target selector for load balancing."""
    
    @abstractmethod
    def select(
        self,
        route: Route,
        context: RoutingContext,
    ) -> RouteTarget:
        pass


class RoundRobinSelector(TargetSelector):
    """Round robin target selector."""
    
    def select(self, route: Route, context: RoutingContext) -> RouteTarget:
        if not route.targets:
            raise NoRouteFoundError("No targets for route")
        
        index = route._rr_index % len(route.targets)
        route._rr_index += 1
        
        return route.targets[index]


class RandomSelector(TargetSelector):
    """Random target selector."""
    
    def select(self, route: Route, context: RoutingContext) -> RouteTarget:
        if not route.targets:
            raise NoRouteFoundError("No targets for route")
        
        return random.choice(route.targets)


class WeightedSelector(TargetSelector):
    """Weighted target selector."""
    
    def select(self, route: Route, context: RoutingContext) -> RouteTarget:
        if not route.targets:
            raise NoRouteFoundError("No targets for route")
        
        total = sum(t.weight for t in route.targets)
        
        if total == 0:
            return random.choice(route.targets)
        
        r = random.randint(1, total)
        cumulative = 0
        
        for target in route.targets:
            cumulative += target.weight
            if r <= cumulative:
                return target
        
        return route.targets[-1]


class LeastLatencySelector(TargetSelector):
    """Least latency target selector."""
    
    def select(self, route: Route, context: RoutingContext) -> RouteTarget:
        if not route.targets:
            raise NoRouteFoundError("No targets for route")
        
        # Filter targets with data
        with_data = [t for t in route.targets if t.request_count > 0]
        
        if not with_data:
            return random.choice(route.targets)
        
        return min(with_data, key=lambda t: t.avg_latency_ms)


# Request router
class RequestRouter:
    """Request router."""
    
    def __init__(self):
        self._routes: Dict[str, Route] = {}
        self._circuit_breakers: Dict[str, CircuitBreaker] = {}
        
        self._matcher = DefaultRouteMatcher()
        
        self._selectors: Dict[LoadBalanceStrategy, TargetSelector] = {
            LoadBalanceStrategy.ROUND_ROBIN: RoundRobinSelector(),
            LoadBalanceStrategy.RANDOM: RandomSelector(),
            LoadBalanceStrategy.WEIGHTED: WeightedSelector(),
            LoadBalanceStrategy.LEAST_LATENCY: LeastLatencySelector(),
        }
        
        self._listeners: List[Callable] = []
        
        # Statistics
        self._total_requests = 0
        self._successful_requests = 0
        self._failed_requests = 0
        self._routing_times: List[float] = []
    
    async def add_route(
        self,
        pattern: str,
        target: str,
        route_type: RouteType = RouteType.PREFIX,
        weight: int = 100,
        **kwargs,
    ) -> Route:
        """Add a route."""
        route = Route(
            pattern=pattern,
            route_type=route_type,
            targets=[RouteTarget(service=target, weight=weight)],
            **kwargs,
        )
        
        self._routes[route.id] = route
        
        # Create circuit breaker for target
        if route.circuit_breaker_enabled:
            self._get_or_create_circuit_breaker(target)
        
        logger.info(f"Route added: {pattern} -> {target}")
        
        return route
    
    async def add_route_with_targets(
        self,
        pattern: str,
        targets: List[Dict[str, Any]],
        route_type: RouteType = RouteType.PREFIX,
        lb_strategy: LoadBalanceStrategy = LoadBalanceStrategy.ROUND_ROBIN,
        **kwargs,
    ) -> Route:
        """Add a route with multiple targets."""
        route_targets = [RouteTarget(**t) for t in targets]
        
        route = Route(
            pattern=pattern,
            route_type=route_type,
            targets=route_targets,
            lb_strategy=lb_strategy,
            **kwargs,
        )
        
        self._routes[route.id] = route
        
        # Create circuit breakers for targets
        if route.circuit_breaker_enabled:
            for target in route_targets:
                self._get_or_create_circuit_breaker(target.service)
        
        logger.info(f"Route added: {pattern} -> {len(targets)} targets")
        
        return route
    
    async def remove_route(self, route_id: str) -> bool:
        """Remove a route."""
        if route_id in self._routes:
            del self._routes[route_id]
            return True
        return False
    
    async def route(
        self,
        path: str,
        method: str = "GET",
        headers: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> RoutingResult:
        """Route a request."""
        start_time = time.monotonic()
        self._total_requests += 1
        
        context = RoutingContext(
            path=path,
            method=method,
            headers=headers or {},
            **kwargs,
        )
        
        # Find matching route
        route = await self._find_route(context)
        
        if not route:
            self._failed_requests += 1
            raise NoRouteFoundError(f"No route found for: {path}")
        
        # Select target
        selector = self._selectors.get(
            route.lb_strategy,
            self._selectors[LoadBalanceStrategy.ROUND_ROBIN],
        )
        
        target = selector.select(route, context)
        
        # Check circuit breaker
        if route.circuit_breaker_enabled:
            cb = self._circuit_breakers.get(target.service)
            
            if cb and cb.is_open:
                # Check if we should transition to half-open
                if cb.opened_at:
                    elapsed = (datetime.utcnow() - cb.opened_at).total_seconds()
                    if elapsed >= cb.timeout_seconds:
                        cb.state = CircuitState.HALF_OPEN
                    else:
                        self._failed_requests += 1
                        raise CircuitOpenError(f"Circuit open for: {target.service}")
        
        # Transform path
        transformed_path = self._transform_path(route, path)
        
        routing_time = (time.monotonic() - start_time) * 1000
        self._routing_times.append(routing_time)
        
        if len(self._routing_times) > 1000:
            self._routing_times = self._routing_times[-1000:]
        
        self._successful_requests += 1
        
        return RoutingResult(
            route=route,
            target=target,
            transformed_path=transformed_path,
            routing_time_ms=routing_time,
        )
    
    async def record_success(self, service: str, latency_ms: float) -> None:
        """Record successful request."""
        cb = self._circuit_breakers.get(service)
        
        if cb:
            cb.success_count += 1
            cb.failure_count = 0
            
            if cb.state == CircuitState.HALF_OPEN:
                if cb.success_count >= cb.success_threshold:
                    cb.state = CircuitState.CLOSED
                    cb.opened_at = None
                    logger.info(f"Circuit closed for: {service}")
        
        # Update target stats
        for route in self._routes.values():
            for target in route.targets:
                if target.service == service:
                    target.request_count += 1
                    target.total_latency_ms += latency_ms
    
    async def record_failure(self, service: str) -> None:
        """Record failed request."""
        cb = self._circuit_breakers.get(service)
        
        if cb:
            cb.failure_count += 1
            cb.success_count = 0
            cb.last_failure = datetime.utcnow()
            
            if cb.failure_count >= cb.failure_threshold:
                cb.state = CircuitState.OPEN
                cb.opened_at = datetime.utcnow()
                logger.warning(f"Circuit opened for: {service}")
        
        # Update target stats
        for route in self._routes.values():
            for target in route.targets:
                if target.service == service:
                    target.request_count += 1
                    target.error_count += 1
    
    def protected_call(self, service: str):
        """Context manager for protected calls."""
        return CircuitBreakerContext(self, service)
    
    async def calculate_retry_delay(
        self,
        attempt: int,
        policy: RetryPolicy,
    ) -> float:
        """Calculate retry delay."""
        if policy.strategy == RetryStrategy.FIXED:
            delay = policy.base_delay_seconds
        
        elif policy.strategy == RetryStrategy.LINEAR:
            delay = policy.base_delay_seconds * attempt
        
        elif policy.strategy == RetryStrategy.EXPONENTIAL:
            delay = policy.base_delay_seconds * (2 ** (attempt - 1))
        
        elif policy.strategy == RetryStrategy.FIBONACCI:
            a, b = 1, 1
            for _ in range(attempt - 1):
                a, b = b, a + b
            delay = policy.base_delay_seconds * a
        
        else:
            delay = policy.base_delay_seconds
        
        # Add jitter
        jitter = delay * random.uniform(0, 0.1)
        delay = delay + jitter
        
        return min(delay, policy.max_delay_seconds)
    
    async def get_circuit_breaker_status(self, service: str) -> Optional[Dict[str, Any]]:
        """Get circuit breaker status."""
        cb = self._circuit_breakers.get(service)
        
        if not cb:
            return None
        
        return {
            "name": cb.name,
            "state": cb.state.value,
            "failure_count": cb.failure_count,
            "success_count": cb.success_count,
            "opened_at": cb.opened_at.isoformat() if cb.opened_at else None,
        }
    
    async def reset_circuit_breaker(self, service: str) -> bool:
        """Reset circuit breaker."""
        cb = self._circuit_breakers.get(service)
        
        if cb:
            cb.state = CircuitState.CLOSED
            cb.failure_count = 0
            cb.success_count = 0
            cb.opened_at = None
            logger.info(f"Circuit breaker reset for: {service}")
            return True
        
        return False
    
    async def get_stats(self) -> RouterStats:
        """Get router statistics."""
        avg_routing_time = (
            sum(self._routing_times) / len(self._routing_times)
            if self._routing_times else 0.0
        )
        
        circuits_open = sum(
            1 for cb in self._circuit_breakers.values()
            if cb.is_open
        )
        
        return RouterStats(
            total_routes=len(self._routes),
            active_routes=sum(1 for r in self._routes.values() if r.enabled),
            total_requests=self._total_requests,
            successful_requests=self._successful_requests,
            failed_requests=self._failed_requests,
            avg_routing_time_ms=avg_routing_time,
            circuits_open=circuits_open,
        )
    
    async def list_routes(self) -> List[Route]:
        """List all routes."""
        return list(self._routes.values())
    
    async def _find_route(self, context: RoutingContext) -> Optional[Route]:
        """Find matching route."""
        matching = [
            route for route in self._routes.values()
            if self._matcher.matches(route, context)
        ]
        
        if not matching:
            return None
        
        # Sort by priority (higher first)
        matching.sort(key=lambda r: r.priority, reverse=True)
        
        return matching[0]
    
    def _transform_path(self, route: Route, path: str) -> str:
        """Transform path based on route config."""
        result = path
        
        if route.strip_prefix and route.route_type == RouteType.PREFIX:
            result = path[len(route.pattern):] or "/"
        
        if route.add_prefix:
            result = route.add_prefix + result
        
        return result
    
    def _get_or_create_circuit_breaker(self, service: str) -> CircuitBreaker:
        """Get or create circuit breaker."""
        if service not in self._circuit_breakers:
            self._circuit_breakers[service] = CircuitBreaker(name=service)
        
        return self._circuit_breakers[service]
    
    def add_listener(self, listener: Callable) -> None:
        """Add event listener."""
        self._listeners.append(listener)


# Circuit breaker context manager
class CircuitBreakerContext:
    """Circuit breaker context manager."""
    
    def __init__(self, router: RequestRouter, service: str):
        self._router = router
        self._service = service
        self._start_time = 0.0
    
    async def __aenter__(self):
        cb = self._router._circuit_breakers.get(self._service)
        
        if cb and cb.is_open:
            if cb.opened_at:
                elapsed = (datetime.utcnow() - cb.opened_at).total_seconds()
                if elapsed >= cb.timeout_seconds:
                    cb.state = CircuitState.HALF_OPEN
                else:
                    raise CircuitOpenError(f"Circuit open for: {self._service}")
        
        self._start_time = time.monotonic()
        return self
    
    async def __aexit__(self, exc_type, exc_val, exc_tb):
        latency = (time.monotonic() - self._start_time) * 1000
        
        if exc_type is None:
            await self._router.record_success(self._service, latency)
        else:
            await self._router.record_failure(self._service)
        
        return False  # Don't suppress exceptions


# Factory functions
def create_request_router() -> RequestRouter:
    """Create request router."""
    return RequestRouter()


def create_route(
    pattern: str,
    target: str,
    **kwargs,
) -> Route:
    """Create a route."""
    return Route(
        pattern=pattern,
        targets=[RouteTarget(service=target)],
        **kwargs,
    )


def create_retry_policy(
    max_retries: int = 3,
    strategy: RetryStrategy = RetryStrategy.EXPONENTIAL,
    **kwargs,
) -> RetryPolicy:
    """Create retry policy."""
    return RetryPolicy(
        max_retries=max_retries,
        strategy=strategy,
        **kwargs,
    )


__all__ = [
    # Exceptions
    "RoutingError",
    "NoRouteFoundError",
    "CircuitOpenError",
    # Enums
    "RouteType",
    "CircuitState",
    "RetryStrategy",
    "LoadBalanceStrategy",
    # Data classes
    "RouteTarget",
    "Route",
    "CircuitBreaker",
    "RetryPolicy",
    "RoutingContext",
    "RoutingResult",
    "RouterStats",
    # Matcher
    "RouteMatcher",
    "DefaultRouteMatcher",
    # Selector
    "TargetSelector",
    "RoundRobinSelector",
    "RandomSelector",
    "WeightedSelector",
    "LeastLatencySelector",
    # Router
    "RequestRouter",
    "CircuitBreakerContext",
    # Factory functions
    "create_request_router",
    "create_route",
    "create_retry_policy",
]



"""
LLM Provider Adapters.

Pre-built adapters for popular LLM providers with auto-configuration
from environment variables.

Supported Providers:
- OpenAI (GPT-4, GPT-4o, GPT-3.5)
- Anthropic (Claude 3.5, Claude 3)
- Google (Gemini 2.0, Gemini 1.5)

Usage:
    >>> from agenticaiframework.llms.providers import OpenAIProvider
    >>> provider = OpenAIProvider.from_env()
    >>> response = provider.generate("Hello!")
    
Or auto-detect from environment:
    >>> from agenticaiframework.llms.providers import auto_detect_provider
    >>> provider = auto_detect_provider()
"""

import os
import logging
from typing import List, Optional, Type

from .base import (
    BaseLLMProvider,
    LLMMessage,
    LLMResponse,
    ProviderConfig,
)
from .openai_provider import OpenAIProvider
from .anthropic_provider import AnthropicProvider
from .google_provider import GoogleProvider

logger = logging.getLogger(__name__)


# Provider detection order
PROVIDER_PRIORITY: List[Type[BaseLLMProvider]] = [
    OpenAIProvider,
    AnthropicProvider,
    GoogleProvider,
]

# Environment variable to provider mapping
ENV_TO_PROVIDER = {
    "OPENAI_API_KEY": OpenAIProvider,
    "ANTHROPIC_API_KEY": AnthropicProvider,
    "GOOGLE_API_KEY": GoogleProvider,
    "GEMINI_API_KEY": GoogleProvider,
}


def auto_detect_provider(
    preferred: Optional[str] = None,
) -> Optional[BaseLLMProvider]:
    """
    Auto-detect and create a provider from environment variables.
    
    Checks for API keys in order:
    1. OPENAI_API_KEY -> OpenAIProvider
    2. ANTHROPIC_API_KEY -> AnthropicProvider  
    3. GOOGLE_API_KEY or GEMINI_API_KEY -> GoogleProvider
    
    Args:
        preferred: Preferred provider name ('openai', 'anthropic', 'google')
        
    Returns:
        Configured provider or None if no API keys found
        
    Example:
        >>> provider = auto_detect_provider()
        >>> if provider:
        ...     response = provider.generate("Hello!")
    """
    # Check preferred first
    if preferred:
        preferred = preferred.lower()
        if preferred == "openai" and os.getenv("OPENAI_API_KEY"):
            return OpenAIProvider.from_env()
        elif preferred == "anthropic" and os.getenv("ANTHROPIC_API_KEY"):
            return AnthropicProvider.from_env()
        elif preferred in ("google", "gemini"):
            if os.getenv("GOOGLE_API_KEY") or os.getenv("GEMINI_API_KEY"):
                return GoogleProvider.from_env()
    
    # Auto-detect from environment
    for env_var, provider_class in ENV_TO_PROVIDER.items():
        if os.getenv(env_var):
            logger.info("Auto-detected %s from %s", provider_class.__name__, env_var)
            return provider_class.from_env()
    
    logger.warning("No LLM provider API keys found in environment")
    return None


def get_provider(
    name: str,
    model: Optional[str] = None,
    api_key: Optional[str] = None,
    **kwargs,
) -> BaseLLMProvider:
    """
    Get a provider by name.
    
    Args:
        name: Provider name ('openai', 'anthropic', 'google')
        model: Default model to use
        api_key: API key (uses env var if not provided)
        **kwargs: Additional provider config
        
    Returns:
        Configured provider instance
        
    Raises:
        ValueError: If provider name is unknown
        
    Example:
        >>> provider = get_provider("openai", model="gpt-4o")
    """
    name = name.lower()
    
    if name == "openai":
        config = ProviderConfig(
            api_key=api_key or os.getenv("OPENAI_API_KEY"),
            default_model=model,
            **kwargs,
        )
        return OpenAIProvider(config)
    
    elif name == "anthropic":
        config = ProviderConfig(
            api_key=api_key or os.getenv("ANTHROPIC_API_KEY"),
            default_model=model,
            **kwargs,
        )
        return AnthropicProvider(config)
    
    elif name in ("google", "gemini"):
        config = ProviderConfig(
            api_key=api_key or os.getenv("GOOGLE_API_KEY") or os.getenv("GEMINI_API_KEY"),
            default_model=model,
            **kwargs,
        )
        return GoogleProvider(config)
    
    else:
        raise ValueError(
            f"Unknown provider: {name}. "
            f"Supported: openai, anthropic, google"
        )


def list_available_providers() -> List[str]:
    """
    List providers that have API keys configured.
    
    Returns:
        List of provider names with valid API keys
        
    Example:
        >>> available = list_available_providers()
        >>> print(available)  # ['openai', 'anthropic']
    """
    available = []
    
    if os.getenv("OPENAI_API_KEY"):
        available.append("openai")
    if os.getenv("ANTHROPIC_API_KEY"):
        available.append("anthropic")
    if os.getenv("GOOGLE_API_KEY") or os.getenv("GEMINI_API_KEY"):
        available.append("google")
    
    return available


__all__ = [
    # Base classes
    "BaseLLMProvider",
    "LLMMessage",
    "LLMResponse",
    "ProviderConfig",
    # Providers
    "OpenAIProvider",
    "AnthropicProvider",
    "GoogleProvider",
    # Factory functions
    "auto_detect_provider",
    "get_provider",
    "list_available_providers",
]



"""
Enterprise API Versioning Module.

API version management, routing, deprecation
tracking, and migration support.

Example:
    # Create versioned API
    api = create_versioned_api("users")
    
    # Register version handlers
    @api.version("v1")
    async def get_users_v1(request):
        return {"users": [...]}
    
    @api.version("v2")
    async def get_users_v2(request):
        return {"data": {"users": [...]}, "meta": {...}}
    
    # Route request
    response = await api.route(request, version="v2")
    
    # Deprecate version
    api.deprecate("v1", sunset_date="2024-12-31")
"""

from __future__ import annotations

import asyncio
import functools
import logging
import re
import uuid
from abc import ABC, abstractmethod
from collections import defaultdict
from dataclasses import dataclass, field
from datetime import datetime, date
from enum import Enum
from typing import (
    Any,
    Awaitable,
    Callable,
    Dict,
    Generic,
    Iterator,
    List,
    Optional,
    Set,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')
RequestHandler = Callable[..., Awaitable[Any]]


logger = logging.getLogger(__name__)


class VersionError(Exception):
    """Version error."""
    pass


class VersionNotFoundError(VersionError):
    """Version not found."""
    pass


class DeprecatedVersionError(VersionError):
    """Deprecated version warning."""
    pass


class SunsetVersionError(VersionError):
    """Version has been sunset."""
    pass


class VersionFormat(str, Enum):
    """Version format types."""
    SEMANTIC = "semantic"  # v1.2.3
    MAJOR = "major"  # v1, v2
    DATE = "date"  # 2024-01-15
    HEADER = "header"  # Via header
    PATH = "path"  # /v1/resource
    QUERY = "query"  # ?version=v1


class VersionStatus(str, Enum):
    """Version status."""
    CURRENT = "current"
    STABLE = "stable"
    DEPRECATED = "deprecated"
    SUNSET = "sunset"
    BETA = "beta"
    ALPHA = "alpha"


class VersionExtractionStrategy(str, Enum):
    """How to extract version from request."""
    URL_PATH = "url_path"
    QUERY_PARAM = "query_param"
    HEADER = "header"
    ACCEPT_HEADER = "accept_header"


@dataclass
class Version:
    """API version."""
    name: str
    status: VersionStatus = VersionStatus.STABLE
    release_date: Optional[date] = None
    deprecation_date: Optional[date] = None
    sunset_date: Optional[date] = None
    description: str = ""
    changelog: List[str] = field(default_factory=list)
    metadata: Dict[str, Any] = field(default_factory=dict)
    
    def is_deprecated(self) -> bool:
        """Check if deprecated."""
        if self.status == VersionStatus.DEPRECATED:
            return True
        
        if self.deprecation_date:
            return date.today() >= self.deprecation_date
        
        return False
    
    def is_sunset(self) -> bool:
        """Check if sunset."""
        if self.status == VersionStatus.SUNSET:
            return True
        
        if self.sunset_date:
            return date.today() >= self.sunset_date
        
        return False


@dataclass
class VersionMapping:
    """Version to handler mapping."""
    version: Version
    handler: RequestHandler
    transformers: List[Callable] = field(default_factory=list)


@dataclass
class RoutingResult:
    """Routing result."""
    version: Version
    handler: RequestHandler
    warnings: List[str] = field(default_factory=list)


@dataclass
class MigrationGuide:
    """Migration guide between versions."""
    from_version: str
    to_version: str
    breaking_changes: List[str] = field(default_factory=list)
    new_features: List[str] = field(default_factory=list)
    removed_features: List[str] = field(default_factory=list)
    migration_steps: List[str] = field(default_factory=list)


@dataclass
class VersionStats:
    """Version statistics."""
    version: str
    request_count: int = 0
    error_count: int = 0
    avg_latency_ms: float = 0.0
    last_accessed: Optional[datetime] = None


# Version parser
class VersionParser:
    """Parse version strings."""
    
    @staticmethod
    def parse(version: str) -> Tuple[int, int, int]:
        """Parse semantic version."""
        # Handle v prefix
        v = version.lstrip("vV")
        
        parts = v.split(".")
        major = int(parts[0]) if len(parts) > 0 else 0
        minor = int(parts[1]) if len(parts) > 1 else 0
        patch = int(parts[2]) if len(parts) > 2 else 0
        
        return (major, minor, patch)
    
    @staticmethod
    def compare(v1: str, v2: str) -> int:
        """Compare versions. Returns -1, 0, or 1."""
        p1 = VersionParser.parse(v1)
        p2 = VersionParser.parse(v2)
        
        if p1 < p2:
            return -1
        elif p1 > p2:
            return 1
        return 0
    
    @staticmethod
    def is_compatible(required: str, actual: str) -> bool:
        """Check if versions are compatible."""
        r = VersionParser.parse(required)
        a = VersionParser.parse(actual)
        
        # Same major version
        return r[0] == a[0]


# Version extractor
class VersionExtractor:
    """Extract version from request."""
    
    def __init__(
        self,
        strategy: VersionExtractionStrategy = VersionExtractionStrategy.URL_PATH,
        header_name: str = "X-API-Version",
        query_param: str = "version",
        path_pattern: str = r"/v(\d+(?:\.\d+)?(?:\.\d+)?)",
        accept_pattern: str = r"application/vnd\.api\.v(\d+)\+json",
    ):
        self._strategy = strategy
        self._header_name = header_name
        self._query_param = query_param
        self._path_pattern = re.compile(path_pattern)
        self._accept_pattern = re.compile(accept_pattern)
    
    def extract(
        self,
        path: str = "",
        headers: Optional[Dict[str, str]] = None,
        query: Optional[Dict[str, str]] = None,
    ) -> Optional[str]:
        """Extract version from request."""
        headers = headers or {}
        query = query or {}
        
        if self._strategy == VersionExtractionStrategy.URL_PATH:
            match = self._path_pattern.search(path)
            if match:
                return f"v{match.group(1)}"
        
        elif self._strategy == VersionExtractionStrategy.HEADER:
            return headers.get(self._header_name)
        
        elif self._strategy == VersionExtractionStrategy.QUERY_PARAM:
            return query.get(self._query_param)
        
        elif self._strategy == VersionExtractionStrategy.ACCEPT_HEADER:
            accept = headers.get("Accept", "")
            match = self._accept_pattern.search(accept)
            if match:
                return f"v{match.group(1)}"
        
        return None


# Response transformer
class ResponseTransformer(ABC):
    """Abstract response transformer."""
    
    @abstractmethod
    async def transform(
        self,
        response: Any,
        from_version: str,
        to_version: str,
    ) -> Any:
        """Transform response between versions."""
        pass


class FieldMappingTransformer(ResponseTransformer):
    """Transform response by mapping fields."""
    
    def __init__(
        self,
        mapping: Dict[str, str],
        remove: Optional[List[str]] = None,
    ):
        self._mapping = mapping
        self._remove = set(remove or [])
    
    async def transform(
        self,
        response: Any,
        from_version: str,
        to_version: str,
    ) -> Any:
        if not isinstance(response, dict):
            return response
        
        result = {}
        
        for key, value in response.items():
            if key in self._remove:
                continue
            
            new_key = self._mapping.get(key, key)
            result[new_key] = value
        
        return result


# Versioned API
class VersionedAPI:
    """
    Versioned API manager.
    """
    
    def __init__(
        self,
        name: str,
        default_version: str = "v1",
        extraction_strategy: VersionExtractionStrategy = VersionExtractionStrategy.URL_PATH,
    ):
        self.name = name
        self._default_version = default_version
        self._versions: Dict[str, Version] = {}
        self._handlers: Dict[str, Dict[str, VersionMapping]] = defaultdict(dict)
        self._transformers: Dict[Tuple[str, str], ResponseTransformer] = {}
        self._migration_guides: Dict[Tuple[str, str], MigrationGuide] = {}
        self._stats: Dict[str, VersionStats] = {}
        self._extractor = VersionExtractor(strategy=extraction_strategy)
    
    def register_version(
        self,
        name: str,
        status: VersionStatus = VersionStatus.STABLE,
        release_date: Optional[date] = None,
        description: str = "",
    ) -> Version:
        """
        Register API version.
        
        Args:
            name: Version name (e.g., "v1")
            status: Version status
            release_date: Release date
            description: Version description
            
        Returns:
            Version object
        """
        version = Version(
            name=name,
            status=status,
            release_date=release_date or date.today(),
            description=description,
        )
        
        self._versions[name] = version
        self._stats[name] = VersionStats(version=name)
        
        return version
    
    def version(
        self,
        version_name: str,
        endpoint: str = "default",
    ) -> Callable:
        """
        Decorator to register versioned handler.
        
        Args:
            version_name: Version name
            endpoint: Endpoint name
            
        Returns:
            Decorator
        """
        def decorator(func: RequestHandler) -> RequestHandler:
            self.register_handler(endpoint, version_name, func)
            return func
        
        return decorator
    
    def register_handler(
        self,
        endpoint: str,
        version_name: str,
        handler: RequestHandler,
    ) -> None:
        """Register handler for endpoint version."""
        if version_name not in self._versions:
            self.register_version(version_name)
        
        version = self._versions[version_name]
        
        self._handlers[endpoint][version_name] = VersionMapping(
            version=version,
            handler=handler,
        )
    
    async def route(
        self,
        endpoint: str,
        version: Optional[str] = None,
        path: str = "",
        headers: Optional[Dict[str, str]] = None,
        query: Optional[Dict[str, str]] = None,
        **kwargs,
    ) -> Any:
        """
        Route request to versioned handler.
        
        Args:
            endpoint: Endpoint name
            version: Explicit version
            path: Request path (for extraction)
            headers: Request headers
            query: Query parameters
            **kwargs: Handler arguments
            
        Returns:
            Handler response
        """
        # Extract version
        if version is None:
            version = self._extractor.extract(path, headers, query)
        
        if version is None:
            version = self._default_version
        
        # Get routing result
        result = self._resolve_handler(endpoint, version)
        
        # Check sunset
        if result.version.is_sunset():
            raise SunsetVersionError(
                f"Version {version} has been sunset"
            )
        
        # Track stats
        self._record_access(version)
        
        # Execute handler
        import time
        start = time.perf_counter()
        
        try:
            response = await result.handler(**kwargs)
            
            latency = (time.perf_counter() - start) * 1000
            self._record_latency(version, latency)
            
            return response
            
        except Exception as e:
            self._record_error(version)
            raise
    
    def _resolve_handler(
        self,
        endpoint: str,
        version: str,
    ) -> RoutingResult:
        """Resolve handler for version."""
        endpoint_handlers = self._handlers.get(endpoint, {})
        
        if version in endpoint_handlers:
            mapping = endpoint_handlers[version]
            warnings = []
            
            if mapping.version.is_deprecated():
                warnings.append(
                    f"Version {version} is deprecated"
                )
            
            return RoutingResult(
                version=mapping.version,
                handler=mapping.handler,
                warnings=warnings,
            )
        
        # Find compatible version
        for v_name, mapping in sorted(
            endpoint_handlers.items(),
            key=lambda x: VersionParser.parse(x[0]),
            reverse=True,
        ):
            if VersionParser.is_compatible(version, v_name):
                return RoutingResult(
                    version=mapping.version,
                    handler=mapping.handler,
                    warnings=[f"Falling back to version {v_name}"],
                )
        
        raise VersionNotFoundError(f"No handler for version {version}")
    
    def deprecate(
        self,
        version: str,
        deprecation_date: Optional[date] = None,
        sunset_date: Optional[date] = None,
    ) -> None:
        """
        Deprecate a version.
        
        Args:
            version: Version to deprecate
            deprecation_date: When deprecated
            sunset_date: When to sunset
        """
        if version in self._versions:
            v = self._versions[version]
            v.status = VersionStatus.DEPRECATED
            v.deprecation_date = deprecation_date or date.today()
            v.sunset_date = sunset_date
    
    def sunset(self, version: str) -> None:
        """Sunset a version."""
        if version in self._versions:
            v = self._versions[version]
            v.status = VersionStatus.SUNSET
            v.sunset_date = date.today()
    
    def add_transformer(
        self,
        from_version: str,
        to_version: str,
        transformer: ResponseTransformer,
    ) -> None:
        """Add response transformer."""
        self._transformers[(from_version, to_version)] = transformer
    
    async def transform_response(
        self,
        response: Any,
        from_version: str,
        to_version: str,
    ) -> Any:
        """Transform response between versions."""
        key = (from_version, to_version)
        
        if key in self._transformers:
            return await self._transformers[key].transform(
                response, from_version, to_version
            )
        
        return response
    
    def add_migration_guide(
        self,
        from_version: str,
        to_version: str,
        guide: MigrationGuide,
    ) -> None:
        """Add migration guide."""
        self._migration_guides[(from_version, to_version)] = guide
    
    def get_migration_guide(
        self,
        from_version: str,
        to_version: str,
    ) -> Optional[MigrationGuide]:
        """Get migration guide."""
        return self._migration_guides.get((from_version, to_version))
    
    def _record_access(self, version: str) -> None:
        """Record version access."""
        if version in self._stats:
            self._stats[version].request_count += 1
            self._stats[version].last_accessed = datetime.utcnow()
    
    def _record_latency(self, version: str, latency_ms: float) -> None:
        """Record request latency."""
        if version in self._stats:
            stats = self._stats[version]
            count = stats.request_count
            
            # Running average
            stats.avg_latency_ms = (
                (stats.avg_latency_ms * (count - 1) + latency_ms) / count
            )
    
    def _record_error(self, version: str) -> None:
        """Record error."""
        if version in self._stats:
            self._stats[version].error_count += 1
    
    def get_versions(self) -> List[Version]:
        """Get all versions."""
        return list(self._versions.values())
    
    def get_version(self, name: str) -> Optional[Version]:
        """Get version by name."""
        return self._versions.get(name)
    
    def get_current_version(self) -> Optional[Version]:
        """Get current (latest stable) version."""
        stable = [
            v for v in self._versions.values()
            if v.status == VersionStatus.STABLE
        ]
        
        if not stable:
            return None
        
        return max(
            stable,
            key=lambda v: VersionParser.parse(v.name)
        )
    
    def get_stats(self, version: Optional[str] = None) -> Union[VersionStats, Dict[str, VersionStats]]:
        """Get version statistics."""
        if version:
            return self._stats.get(version, VersionStats(version=version))
        return dict(self._stats)


# Decorator for simple versioning
def versioned(
    api: VersionedAPI,
    endpoint: str = "default",
):
    """
    Create versioned endpoint decorator.
    
    Args:
        api: Versioned API
        endpoint: Endpoint name
    """
    def version_decorator(version_name: str):
        def decorator(func: RequestHandler) -> RequestHandler:
            api.register_handler(endpoint, version_name, func)
            return func
        return decorator
    
    return version_decorator


# Factory functions
def create_versioned_api(
    name: str,
    default_version: str = "v1",
    extraction_strategy: VersionExtractionStrategy = VersionExtractionStrategy.URL_PATH,
) -> VersionedAPI:
    """Create versioned API."""
    return VersionedAPI(
        name=name,
        default_version=default_version,
        extraction_strategy=extraction_strategy,
    )


def create_version_extractor(
    strategy: VersionExtractionStrategy = VersionExtractionStrategy.URL_PATH,
    **kwargs,
) -> VersionExtractor:
    """Create version extractor."""
    return VersionExtractor(strategy=strategy, **kwargs)


def create_migration_guide(
    from_version: str,
    to_version: str,
    breaking_changes: Optional[List[str]] = None,
    new_features: Optional[List[str]] = None,
) -> MigrationGuide:
    """Create migration guide."""
    return MigrationGuide(
        from_version=from_version,
        to_version=to_version,
        breaking_changes=breaking_changes or [],
        new_features=new_features or [],
    )


__all__ = [
    # Exceptions
    "VersionError",
    "VersionNotFoundError",
    "DeprecatedVersionError",
    "SunsetVersionError",
    # Enums
    "VersionFormat",
    "VersionStatus",
    "VersionExtractionStrategy",
    # Data classes
    "Version",
    "VersionMapping",
    "RoutingResult",
    "MigrationGuide",
    "VersionStats",
    # Parser
    "VersionParser",
    # Extractor
    "VersionExtractor",
    # Transformer
    "ResponseTransformer",
    "FieldMappingTransformer",
    # Main class
    "VersionedAPI",
    # Decorators
    "versioned",
    # Factory functions
    "create_versioned_api",
    "create_version_extractor",
    "create_migration_guide",
]



"""
Communication manager for protocol-based message routing.

Thread-safe protocol registration and message dispatch.
"""

from __future__ import annotations

import logging
import threading
from typing import Any, Callable

from .exceptions import ProtocolError, ProtocolNotFoundError  # noqa: F401 - exported for library users

logger = logging.getLogger(__name__)


class CommunicationManager:
    """Thread-safe communication manager."""

    def __init__(self) -> None:
        self._lock = threading.Lock()
        self.protocols: dict[str, Callable[[Any], Any]] = {}

    def register_protocol(self, name: str, handler_fn: Callable[[Any], Any]) -> None:
        with self._lock:
            self.protocols[name] = handler_fn
        logger.info("[CommunicationManager] Registered protocol '%s'", name)

    def register_handler(self, handler_fn: Callable[[Any], Any], name: str | None = None) -> None:
        """Alternative method for registering handlers."""
        protocol_name = name or f"handler_{len(self.protocols)}"
        self.register_protocol(protocol_name, handler_fn)

    def send(self, protocol: str, data: Any) -> Any:
        with self._lock:
            handler = self.protocols.get(protocol)
        if handler:
            try:
                return handler(data)
            except (TypeError, ValueError, ConnectionError, TimeoutError) as e:
                logger.warning("Protocol '%s' communication failed: %s", protocol, e)
            except Exception:  # noqa: BLE001
                logger.exception("Unexpected error in protocol '%s'", protocol)
        else:
            logger.warning("[CommunicationManager] Protocol '%s' not found", protocol)
        return None

    def list_protocols(self) -> list[str]:
        with self._lock:
            return list(self.protocols.keys())

    def send_message(self, message: Any, protocol: str | None = None) -> Any:
        """Send a message using the specified or first available protocol."""
        if protocol:
            return self.send(protocol, message)
        with self._lock:
            first = next(iter(self.protocols), None)
        if first:
            return self.send(first, message)
        logger.warning("[CommunicationManager] No protocols available")
        return None



"""
Enterprise Image Processor Module.

Provides image manipulation including resize, crop,
watermark, format conversion, and compression.

Example:
    # Create image processor
    images = create_image_processor()
    
    # Resize image
    result = await images.resize(image_bytes, width=800, height=600)
    
    # Add watermark
    result = await images.watermark(
        image_bytes,
        text="© 2024 Company",
        position="bottom-right",
    )
    
    # Create thumbnail
    result = await images.thumbnail(image_bytes, size=200)
    
    # Pipeline processing
    result = await (
        images.pipeline(image_bytes)
        .resize(800, 600)
        .crop(100, 100, 500, 400)
        .watermark("© Company")
        .compress(quality=80)
        .execute()
    )
"""

from __future__ import annotations

import asyncio
import base64
import functools
import io
import json
import logging
import uuid
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import (
    Any,
    Callable,
    Dict,
    List,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

T = TypeVar('T')


logger = logging.getLogger(__name__)


class ImageError(Exception):
    """Image processing error."""
    pass


class FormatError(ImageError):
    """Format error."""
    pass


class ProcessingError(ImageError):
    """Processing error."""
    pass


class ImageFormat(str, Enum):
    """Image formats."""
    JPEG = "jpeg"
    PNG = "png"
    WEBP = "webp"
    GIF = "gif"
    BMP = "bmp"
    TIFF = "tiff"
    ICO = "ico"
    AVIF = "avif"


class ResizeMode(str, Enum):
    """Resize modes."""
    FIT = "fit"          # Fit within bounds
    FILL = "fill"        # Fill bounds (may crop)
    STRETCH = "stretch"  # Stretch to exact size
    CONTAIN = "contain"  # Contain with padding
    COVER = "cover"      # Cover bounds


class Position(str, Enum):
    """Position for overlays."""
    CENTER = "center"
    TOP_LEFT = "top-left"
    TOP_RIGHT = "top-right"
    BOTTOM_LEFT = "bottom-left"
    BOTTOM_RIGHT = "bottom-right"
    TOP = "top"
    BOTTOM = "bottom"
    LEFT = "left"
    RIGHT = "right"


class FlipDirection(str, Enum):
    """Flip directions."""
    HORIZONTAL = "horizontal"
    VERTICAL = "vertical"
    BOTH = "both"


@dataclass
class ImageInfo:
    """Image information."""
    width: int = 0
    height: int = 0
    format: ImageFormat = ImageFormat.JPEG
    mode: str = "RGB"  # RGB, RGBA, L, etc.
    size_bytes: int = 0
    has_alpha: bool = False
    is_animated: bool = False
    frame_count: int = 1
    exif: Dict[str, Any] = field(default_factory=dict)


@dataclass
class ResizeOptions:
    """Resize options."""
    width: Optional[int] = None
    height: Optional[int] = None
    mode: ResizeMode = ResizeMode.FIT
    upscale: bool = False
    filter: str = "lanczos"  # nearest, bilinear, bicubic, lanczos


@dataclass
class CropOptions:
    """Crop options."""
    x: int = 0
    y: int = 0
    width: int = 0
    height: int = 0
    # Alternative: aspect ratio crop
    aspect_ratio: Optional[str] = None  # "16:9", "4:3", "1:1"
    gravity: Position = Position.CENTER


@dataclass
class WatermarkOptions:
    """Watermark options."""
    text: Optional[str] = None
    image_path: Optional[str] = None
    image_data: Optional[bytes] = None
    position: Position = Position.BOTTOM_RIGHT
    opacity: float = 0.5
    font_size: int = 24
    font_color: str = "#FFFFFF"
    padding: int = 10
    rotation: int = 0


@dataclass
class CompressionOptions:
    """Compression options."""
    quality: int = 85
    optimize: bool = True
    progressive: bool = False
    strip_metadata: bool = False


@dataclass
class FilterOptions:
    """Image filter options."""
    brightness: float = 1.0
    contrast: float = 1.0
    saturation: float = 1.0
    sharpness: float = 1.0
    blur: float = 0.0
    grayscale: bool = False
    sepia: bool = False
    invert: bool = False


@dataclass
class ProcessedImage:
    """Processed image result."""
    id: str = field(default_factory=lambda: str(uuid.uuid4()))
    content: bytes = b""
    format: ImageFormat = ImageFormat.JPEG
    width: int = 0
    height: int = 0
    size_bytes: int = 0
    operations: List[str] = field(default_factory=list)
    processed_at: datetime = field(default_factory=datetime.utcnow)


# Image Backend
class ImageBackend(ABC):
    """Abstract image processing backend."""
    
    @abstractmethod
    async def get_info(self, data: bytes) -> ImageInfo:
        """Get image information."""
        pass
    
    @abstractmethod
    async def resize(
        self,
        data: bytes,
        options: ResizeOptions,
    ) -> bytes:
        """Resize image."""
        pass
    
    @abstractmethod
    async def crop(
        self,
        data: bytes,
        options: CropOptions,
    ) -> bytes:
        """Crop image."""
        pass
    
    @abstractmethod
    async def watermark(
        self,
        data: bytes,
        options: WatermarkOptions,
    ) -> bytes:
        """Add watermark."""
        pass
    
    @abstractmethod
    async def convert(
        self,
        data: bytes,
        target_format: ImageFormat,
        options: Optional[CompressionOptions] = None,
    ) -> bytes:
        """Convert format."""
        pass
    
    @abstractmethod
    async def apply_filters(
        self,
        data: bytes,
        options: FilterOptions,
    ) -> bytes:
        """Apply filters."""
        pass


class MockImageBackend(ImageBackend):
    """Mock image backend for testing."""
    
    async def get_info(self, data: bytes) -> ImageInfo:
        """Get mock image info."""
        return ImageInfo(
            width=800,
            height=600,
            format=ImageFormat.JPEG,
            mode="RGB",
            size_bytes=len(data),
        )
    
    async def resize(
        self,
        data: bytes,
        options: ResizeOptions,
    ) -> bytes:
        """Mock resize."""
        info = {
            "operation": "resize",
            "width": options.width,
            "height": options.height,
            "mode": options.mode.value,
            "original_size": len(data),
        }
        return json.dumps(info).encode()
    
    async def crop(
        self,
        data: bytes,
        options: CropOptions,
    ) -> bytes:
        """Mock crop."""
        info = {
            "operation": "crop",
            "x": options.x,
            "y": options.y,
            "width": options.width,
            "height": options.height,
        }
        return json.dumps(info).encode()
    
    async def watermark(
        self,
        data: bytes,
        options: WatermarkOptions,
    ) -> bytes:
        """Mock watermark."""
        info = {
            "operation": "watermark",
            "text": options.text,
            "position": options.position.value,
            "opacity": options.opacity,
        }
        return json.dumps(info).encode()
    
    async def convert(
        self,
        data: bytes,
        target_format: ImageFormat,
        options: Optional[CompressionOptions] = None,
    ) -> bytes:
        """Mock convert."""
        info = {
            "operation": "convert",
            "target_format": target_format.value,
            "quality": options.quality if options else 85,
        }
        return json.dumps(info).encode()
    
    async def apply_filters(
        self,
        data: bytes,
        options: FilterOptions,
    ) -> bytes:
        """Mock apply filters."""
        info = {
            "operation": "filter",
            "brightness": options.brightness,
            "contrast": options.contrast,
            "grayscale": options.grayscale,
        }
        return json.dumps(info).encode()


class ImagePipeline:
    """Fluent image processing pipeline."""
    
    def __init__(
        self,
        processor: "ImageProcessor",
        data: bytes,
    ):
        self._processor = processor
        self._data = data
        self._operations: List[Tuple[str, Any]] = []
    
    def resize(
        self,
        width: Optional[int] = None,
        height: Optional[int] = None,
        mode: ResizeMode = ResizeMode.FIT,
    ) -> "ImagePipeline":
        """Add resize operation."""
        self._operations.append((
            "resize",
            ResizeOptions(width=width, height=height, mode=mode),
        ))
        return self
    
    def crop(
        self,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> "ImagePipeline":
        """Add crop operation."""
        self._operations.append((
            "crop",
            CropOptions(x=x, y=y, width=width, height=height),
        ))
        return self
    
    def watermark(
        self,
        text: str,
        position: Position = Position.BOTTOM_RIGHT,
        opacity: float = 0.5,
    ) -> "ImagePipeline":
        """Add watermark operation."""
        self._operations.append((
            "watermark",
            WatermarkOptions(text=text, position=position, opacity=opacity),
        ))
        return self
    
    def convert(
        self,
        format: ImageFormat,
    ) -> "ImagePipeline":
        """Add format conversion."""
        self._operations.append(("convert", format))
        return self
    
    def compress(
        self,
        quality: int = 85,
    ) -> "ImagePipeline":
        """Add compression."""
        self._operations.append((
            "compress",
            CompressionOptions(quality=quality),
        ))
        return self
    
    def filter(
        self,
        brightness: float = 1.0,
        contrast: float = 1.0,
        saturation: float = 1.0,
        grayscale: bool = False,
    ) -> "ImagePipeline":
        """Add filter operation."""
        self._operations.append((
            "filter",
            FilterOptions(
                brightness=brightness,
                contrast=contrast,
                saturation=saturation,
                grayscale=grayscale,
            ),
        ))
        return self
    
    def rotate(self, degrees: int) -> "ImagePipeline":
        """Add rotation."""
        self._operations.append(("rotate", degrees))
        return self
    
    def flip(self, direction: FlipDirection) -> "ImagePipeline":
        """Add flip operation."""
        self._operations.append(("flip", direction))
        return self
    
    async def execute(self) -> ProcessedImage:
        """Execute pipeline."""
        data = self._data
        operations = []
        
        for op_name, op_args in self._operations:
            if op_name == "resize":
                data = await self._processor.backend.resize(data, op_args)
            elif op_name == "crop":
                data = await self._processor.backend.crop(data, op_args)
            elif op_name == "watermark":
                data = await self._processor.backend.watermark(data, op_args)
            elif op_name == "convert":
                data = await self._processor.backend.convert(data, op_args)
            elif op_name == "compress":
                data = await self._processor.backend.convert(
                    data, ImageFormat.JPEG, op_args
                )
            elif op_name == "filter":
                data = await self._processor.backend.apply_filters(data, op_args)
            
            operations.append(op_name)
        
        return ProcessedImage(
            content=data,
            size_bytes=len(data),
            operations=operations,
        )


class ImageProcessor:
    """
    Image processing service.
    """
    
    def __init__(
        self,
        backend: Optional[ImageBackend] = None,
    ):
        self._backend = backend or MockImageBackend()
    
    @property
    def backend(self) -> ImageBackend:
        """Get backend."""
        return self._backend
    
    def pipeline(self, data: bytes) -> ImagePipeline:
        """Create processing pipeline."""
        return ImagePipeline(self, data)
    
    async def get_info(self, data: bytes) -> ImageInfo:
        """Get image information."""
        return await self._backend.get_info(data)
    
    async def resize(
        self,
        data: bytes,
        width: Optional[int] = None,
        height: Optional[int] = None,
        mode: ResizeMode = ResizeMode.FIT,
        **kwargs,
    ) -> ProcessedImage:
        """
        Resize image.
        
        Args:
            data: Image bytes
            width: Target width
            height: Target height
            mode: Resize mode
            
        Returns:
            Processed image
        """
        options = ResizeOptions(
            width=width,
            height=height,
            mode=mode,
            **kwargs,
        )
        
        content = await self._backend.resize(data, options)
        
        return ProcessedImage(
            content=content,
            width=width or 0,
            height=height or 0,
            size_bytes=len(content),
            operations=["resize"],
        )
    
    async def thumbnail(
        self,
        data: bytes,
        size: int = 150,
        mode: ResizeMode = ResizeMode.FIT,
    ) -> ProcessedImage:
        """Create thumbnail."""
        return await self.resize(data, width=size, height=size, mode=mode)
    
    async def crop(
        self,
        data: bytes,
        x: int,
        y: int,
        width: int,
        height: int,
    ) -> ProcessedImage:
        """
        Crop image.
        
        Args:
            data: Image bytes
            x: Left position
            y: Top position
            width: Crop width
            height: Crop height
            
        Returns:
            Processed image
        """
        options = CropOptions(x=x, y=y, width=width, height=height)
        content = await self._backend.crop(data, options)
        
        return ProcessedImage(
            content=content,
            width=width,
            height=height,
            size_bytes=len(content),
            operations=["crop"],
        )
    
    async def crop_aspect(
        self,
        data: bytes,
        aspect_ratio: str = "1:1",
        gravity: Position = Position.CENTER,
    ) -> ProcessedImage:
        """Crop to aspect ratio."""
        options = CropOptions(aspect_ratio=aspect_ratio, gravity=gravity)
        content = await self._backend.crop(data, options)
        
        return ProcessedImage(
            content=content,
            size_bytes=len(content),
            operations=["crop_aspect"],
        )
    
    async def watermark(
        self,
        data: bytes,
        text: Optional[str] = None,
        image_path: Optional[str] = None,
        image_data: Optional[bytes] = None,
        position: Position = Position.BOTTOM_RIGHT,
        opacity: float = 0.5,
        **kwargs,
    ) -> ProcessedImage:
        """
        Add watermark.
        
        Args:
            data: Image bytes
            text: Watermark text
            image_path: Watermark image path
            image_data: Watermark image bytes
            position: Watermark position
            opacity: Watermark opacity
            
        Returns:
            Processed image
        """
        options = WatermarkOptions(
            text=text,
            image_path=image_path,
            image_data=image_data,
            position=position,
            opacity=opacity,
            **kwargs,
        )
        
        content = await self._backend.watermark(data, options)
        
        return ProcessedImage(
            content=content,
            size_bytes=len(content),
            operations=["watermark"],
        )
    
    async def convert(
        self,
        data: bytes,
        target_format: ImageFormat,
        quality: int = 85,
    ) -> ProcessedImage:
        """
        Convert image format.
        
        Args:
            data: Image bytes
            target_format: Target format
            quality: Compression quality
            
        Returns:
            Processed image
        """
        options = CompressionOptions(quality=quality)
        content = await self._backend.convert(data, target_format, options)
        
        return ProcessedImage(
            content=content,
            format=target_format,
            size_bytes=len(content),
            operations=["convert"],
        )
    
    async def compress(
        self,
        data: bytes,
        quality: int = 85,
        format: Optional[ImageFormat] = None,
    ) -> ProcessedImage:
        """
        Compress image.
        
        Args:
            data: Image bytes
            quality: Compression quality (1-100)
            format: Target format
            
        Returns:
            Processed image
        """
        target = format or ImageFormat.JPEG
        options = CompressionOptions(quality=quality, optimize=True)
        content = await self._backend.convert(data, target, options)
        
        return ProcessedImage(
            content=content,
            format=target,
            size_bytes=len(content),
            operations=["compress"],
        )
    
    async def apply_filter(
        self,
        data: bytes,
        brightness: float = 1.0,
        contrast: float = 1.0,
        saturation: float = 1.0,
        grayscale: bool = False,
        **kwargs,
    ) -> ProcessedImage:
        """
        Apply filters.
        
        Args:
            data: Image bytes
            brightness: Brightness adjustment
            contrast: Contrast adjustment
            saturation: Saturation adjustment
            grayscale: Convert to grayscale
            
        Returns:
            Processed image
        """
        options = FilterOptions(
            brightness=brightness,
            contrast=contrast,
            saturation=saturation,
            grayscale=grayscale,
            **kwargs,
        )
        
        content = await self._backend.apply_filters(data, options)
        
        return ProcessedImage(
            content=content,
            size_bytes=len(content),
            operations=["filter"],
        )
    
    async def optimize(
        self,
        data: bytes,
        max_width: int = 1920,
        max_height: int = 1080,
        quality: int = 80,
        format: ImageFormat = ImageFormat.WEBP,
    ) -> ProcessedImage:
        """
        Optimize image for web.
        
        Args:
            data: Image bytes
            max_width: Maximum width
            max_height: Maximum height
            quality: Compression quality
            format: Target format (WEBP recommended)
            
        Returns:
            Optimized image
        """
        # Get info
        info = await self.get_info(data)
        
        result = data
        operations = []
        
        # Resize if needed
        if info.width > max_width or info.height > max_height:
            resize_result = await self.resize(
                result,
                width=max_width,
                height=max_height,
                mode=ResizeMode.FIT,
            )
            result = resize_result.content
            operations.append("resize")
        
        # Convert and compress
        convert_result = await self.convert(result, format, quality)
        result = convert_result.content
        operations.append("convert")
        
        return ProcessedImage(
            content=result,
            format=format,
            size_bytes=len(result),
            operations=operations,
        )
    
    async def save(
        self,
        image: ProcessedImage,
        path: Union[str, Path],
    ) -> Path:
        """Save processed image."""
        path = Path(path)
        path.write_bytes(image.content)
        return path
    
    def to_base64(self, image: ProcessedImage) -> str:
        """Convert to base64."""
        return base64.b64encode(image.content).decode()
    
    def to_data_uri(self, image: ProcessedImage) -> str:
        """Convert to data URI."""
        b64 = self.to_base64(image)
        mime = f"image/{image.format.value}"
        return f"data:{mime};base64,{b64}"


# Decorators
def image_response(
    format: ImageFormat = ImageFormat.JPEG,
) -> Callable:
    """Decorator for image response."""
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        async def wrapper(*args, **kwargs):
            result = await func(*args, **kwargs)
            
            if isinstance(result, ProcessedImage):
                return {
                    "content": result.content,
                    "content_type": f"image/{result.format.value}",
                }
            
            return result
        
        return wrapper
    return decorator


# Factory functions
def create_image_processor(
    backend: Optional[ImageBackend] = None,
) -> ImageProcessor:
    """Create image processor."""
    return ImageProcessor(backend)


def create_resize_options(
    width: Optional[int] = None,
    height: Optional[int] = None,
    mode: ResizeMode = ResizeMode.FIT,
    **kwargs,
) -> ResizeOptions:
    """Create resize options."""
    return ResizeOptions(width=width, height=height, mode=mode, **kwargs)


def create_watermark_options(
    text: Optional[str] = None,
    position: Position = Position.BOTTOM_RIGHT,
    opacity: float = 0.5,
    **kwargs,
) -> WatermarkOptions:
    """Create watermark options."""
    return WatermarkOptions(
        text=text,
        position=position,
        opacity=opacity,
        **kwargs,
    )


def create_filter_options(
    brightness: float = 1.0,
    contrast: float = 1.0,
    saturation: float = 1.0,
    **kwargs,
) -> FilterOptions:
    """Create filter options."""
    return FilterOptions(
        brightness=brightness,
        contrast=contrast,
        saturation=saturation,
        **kwargs,
    )


__all__ = [
    # Exceptions
    "ImageError",
    "FormatError",
    "ProcessingError",
    # Enums
    "ImageFormat",
    "ResizeMode",
    "Position",
    "FlipDirection",
    # Data classes
    "ImageInfo",
    "ResizeOptions",
    "CropOptions",
    "WatermarkOptions",
    "CompressionOptions",
    "FilterOptions",
    "ProcessedImage",
    # Backend
    "ImageBackend",
    "MockImageBackend",
    # Pipeline
    "ImagePipeline",
    # Processor
    "ImageProcessor",
    # Decorators
    "image_response",
    # Factory functions
    "create_image_processor",
    "create_resize_options",
    "create_watermark_options",
    "create_filter_options",
]
